/* my_malloc.c : memory allocation routines with error checking
**
** Written and Copyright (C) 1994 by Michael J. Gourlay
**
** NO WARRANTEES, EXPRESS OR IMPLIED.
*/

#ifdef STHREADS
#include "stfix.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include "my_malloc.h"

#ifdef NOTDEF
char *my_malloc(unsigned size, char *proc)
{
  char *mem;
#ifdef SUNOS
  malloc_verify();
#endif

#ifdef STHREADS
  if((mem=malloc(size))==NULL) {
    fprintf(stderr, "Bad Alloc: %li in %s\n", size, proc);
  }
#else  /* STHREADS */
  if((mem=MM_ALLOC(size))==NULL) {
    fprintf(stderr, "Bad Alloc: %li in %s\n", size, proc);
  }
#endif /* STHREADS */

#ifdef SUNOS
  malloc_verify();
#endif

#if MTCAMP_MEM_TRACE
  RD_MEM_ALLOCED((void *)mem, size);
#endif

  return(mem);
}
#endif


#define EXTRA (nelem)

char *my_calloc(unsigned nelem, unsigned elsize, char *proc)
{
  char *mem;
#ifdef SUNOS
  malloc_verify();
#endif
#ifdef STHREADS
  if((mem=MM_ALLOC((nelem + EXTRA) * elsize))==NULL) {
    fprintf(stderr, "Bad Alloc: %li x %li in %s\n", nelem, elsize, proc);
  }
#else  /* STHREADS */
  if((mem=calloc(nelem + EXTRA, elsize))==NULL) {
    fprintf(stderr, "Bad Alloc: %li x %li in %s\n", nelem, elsize, proc);
  }
#endif /* STHREADS */

#ifdef DEBUG_ALLOC
  printf("allocated %li at %p from %s\n", nelem*elsize, mem, proc);
  fflush(stdout);
#endif

#ifdef SUNOS
  malloc_verify();
#endif
#if MTCAMP_MEM_TRACE
  RD_MEM_ALLOCED((void *)mem, (nelem + EXTRA) * elsize);
#endif
  return(mem);
}

void my_free(void *ptr, char *proc)
{
#ifdef SUNOS
  malloc_verify();
#endif

#ifdef DEBUG_ALLOC
  printf("freeing %p from %s\n", ptr, proc);
  fflush(stdout);
#endif

#ifdef STHREADS
  MM_FREE(ptr);
#else
  free(ptr);
#endif

#ifdef SUNOS
  malloc_verify();
#endif
#if MTCAMP_MEM_TRACE
  RD_MEM_FREED((void *)ptr);
#endif
  return;
}
