/* 
 * racedtct.h --- declarations related for race detection
 */
#define MEM_TRACE_PUT_ID 1

typedef unsigned long source_id_t;

#ifdef __cplusplus
extern "C" {
#endif

#if MEM_TRACE_PUT_ID
  void __rd_raw_fork_begin(source_id_t, char *, int);
#else  /* MEM_TRACE_PUT_ID */
  void __rd_raw_fork_begin(char *, int);
#endif /* MEM_TRACE_PUT_ID */
  void __rd_raw_fork_end();
  void __rd_sync_begin();
  void __rd_sync_end(char *, int);
  void __rd_do_sync();

#if MEM_TRACE_PUT_ID
  void __mem_trace_read(void * addr, int size, source_id_t, char *, int);
  void __mem_trace_write(void * addr, int size, source_id_t, char *, int);
  void __mem_trace_read_write(void * addr, int size, source_id_t, char *, int);
#else  /* MEM_TRACE_PUT_ID */
  void __mem_trace_read(void * addr, int size, char *, int);
  void __mem_trace_write(void * addr, int size, char *, int);
  void __mem_trace_read_write(void * addr, int size, char *, int);
#endif /* MEM_TRACE_PUT_ID */
  void __rd_mem_alloced(void * addr, int size, char * file, int line);
  void __rd_mem_freed(void * addr, char * file, int line);

  void __rd_race_detect_on();
  void __rd_race_detect_off();
#ifdef __cplusplus
}
#endif

#if MEM_TRACE_PUT_ID

#define MEM_TRACE_READ(addr, size, id) \
__mem_trace_read((char*)addr, size, id, __FILE__, __LINE__)
#define MEM_TRACE_WRITE(addr, size, id) \
     __mem_trace_write((char*)addr, size, id, __FILE__, __LINE__)
#define MEM_TRACE_READ_WRITE(addr, size, id) \
     __mem_trace_read_write((char*)addr, size, id, __FILE__, __LINE__)

#define MEM_TRACE_RAW_FORK_BEGIN(id) \
     __rd_raw_fork_begin(id, __FILE__, __LINE__) 

#else

#define MEM_TRACE_READ(addr, size) \
__mem_trace_read((char*)addr, size, __FILE__, __LINE__)
#define MEM_TRACE_WRITE(addr, size) \
     __mem_trace_write((char*)addr, size, __FILE__, __LINE__)
#define MEM_TRACE_READ_WRITE(addr, size) \
     __mem_trace_read_write((char*)addr, size, __FILE__, __LINE__)

#define MEM_TRACE_RAW_FORK_BEGIN() __rd_raw_fork_begin(__FILE__, __LINE__) 

#endif

#define MEM_TRACE_RAW_FORK_END() __rd_raw_fork_end() 
#define MEM_TRACE_SYNC_BEGIN() __rd_sync_begin() 
#define MEM_TRACE_SYNC_END() __rd_sync_end(__FILE__, __LINE__) 
#define MEM_TRACE_DO_SYNC() __rd_do_sync()

#define RD_MEM_ALLOCED(addr, size) __rd_mem_alloced(addr, size, __FILE__, __LINE__)
#define RD_MEM_FREED(addr) __rd_mem_freed(addr, __FILE__, __LINE__)
