/*
 * Copyright 1988, 1989 Hans-J. Boehm, Alan J. Demers
 * Copyright (c) 1991-1996 by Xerox Corporation.  All rights reserved.
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.

 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * This program was derived from a program originally developed by
 * Hans-J. Boehm and Alan J. Demers. Extensions have been made by Toshio
 * Endo.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 */
/* Endo, Oct 16, 1998 */

# define I_HIDE_POINTERS
# include "gc_priv.h"
/* # include "gc_mark.h" */

# ifdef SRC_M3
void GC_push_finalizer_structures(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    GC_push_all(SGC_MYENV_ARG1 (ptr_t)(&GC_dl_head), (ptr_t)(&GC_dl_head) + sizeof(word));
    GC_push_all(SGC_MYENV_ARG1 (ptr_t)(&GC_fo_head), (ptr_t)(&GC_fo_head) + sizeof(word));
}
# endif

/* Double the size of a hash table. *size_ptr is the log of its current	*/
/* size.  May be a noop.						*/
/* *table is a pointer to an array of hash headers.  If we succeed, we	*/
/* update both *table and *log_size_ptr.				*/
/* Lock is held.  Signals are disabled.					*/
void GC_grow_table(SGC_MYENV_ARG1 table, log_size_ptr)
SGC_MYENV_ARG_DECL2
struct hash_chain_entry ***table;
signed_word * log_size_ptr;
{
    register word i;
    register struct hash_chain_entry *p;
    int log_old_size = *log_size_ptr;
    register int log_new_size = log_old_size + 1;
    word old_size = ((log_old_size == -1)? 0: (1 << log_old_size));
    register word new_size = 1 << log_new_size;
    struct hash_chain_entry **new_table = (struct hash_chain_entry **)
    	GC_generic_malloc_inner_ignore_off_page(SGC_MYENV_ARG1 
    		(size_t)new_size * sizeof(struct hash_chain_entry *), /* KAMO NORMAL */ UNCOLLECTABLE);
    
    if (new_table == 0) {
    	if (table == 0) {
    	    ABORT("Insufficient space for initial table allocation");
    	} else {
    	    return;
    	}
    }
    for (i = 0; i < old_size; i++) {
      p = (*table)[i];
      while (p != 0) {
        register ptr_t real_key = (ptr_t)REVEAL_POINTER(p -> hidden_key);
        register struct hash_chain_entry *next = p -> next;
        register int new_hash = HASH3(real_key, new_size, log_new_size);
        
        p -> next = new_table[new_hash];
        new_table[new_hash] = p;
        p = next;
      }
    }
    *log_size_ptr = log_new_size;
    if(*table != 0){
	GC_free(*table);
    }
    *table = new_table;
}

# if defined(__STDC__) || defined(__cplusplus)
    int GC_register_disappearing_link(GC_PTR * link)
# else
    int GC_register_disappearing_link(link)
    GC_PTR * link;
# endif
{
    ptr_t base;
    
    base = (ptr_t)GC_base((GC_PTR)link);
    if (base == 0)
    	ABORT("Bad arg to GC_register_disappearing_link");
    return(GC_general_register_disappearing_link(link, base));
}

# if defined(__STDC__) || defined(__cplusplus)
    int GC_general_register_disappearing_link(GC_PTR * link,
    					      GC_PTR obj)
# else
    int GC_general_register_disappearing_link(link, obj)
    GC_PTR * link;
    GC_PTR obj;
# endif

{
    DECLARE_MYENV3
    struct disappearing_link *curr_dl;
    int index;
    struct disappearing_link * new_dl;
    
    if ((word)link & (ALIGNMENT-1))
    	ABORT("Bad arg to GC_general_register_disappearing_link");
    PSM_mutex_lock_polling_inner(SGC_MYENV_ARG1 &GC_g_dl_ml, SGC_polling_inner);
    if (GC_g_log_dl_table_size == -1
        || GC_g_dl_entries > ((word)1 << GC_g_log_dl_table_size)) {
    	GC_grow_table(SGC_MYENV_ARG1 (struct hash_chain_entry ***)(&GC_g_dl_head),
    		      &GC_g_log_dl_table_size);
#	ifdef PRINTSTATS
	    GC_printf1("Grew dl table to %lu entries\n",
	    		(unsigned long)(1 << GC_g_log_dl_table_size));
#	endif
    }
    index = HASH2(link, GC_g_log_dl_table_size);
    curr_dl = GC_g_dl_head[index];
    for (curr_dl = GC_g_dl_head[index]; curr_dl != 0; curr_dl = dl_next(curr_dl)) {
        if (curr_dl -> dl_hidden_link == HIDE_POINTER(link)) {
            curr_dl -> dl_hidden_obj = HIDE_POINTER(obj);
	    PSM_mutex_unlock(&GC_g_dl_ml);
            return(1);
        }
    }
      new_dl = (struct disappearing_link *)
    	GC_generic_malloc_inner(SGC_MYENV_ARG1 sizeof(struct disappearing_link),NORMAL);
    if (new_dl != 0) {
        new_dl -> dl_hidden_obj = HIDE_POINTER(obj);
        new_dl -> dl_hidden_link = HIDE_POINTER(link);
        dl_set_next(new_dl, GC_g_dl_head[index]);
        GC_g_dl_head[index] = new_dl;
        GC_g_dl_entries++;
    } else {
        GC_g_finalization_failures++;
    }
    PSM_mutex_unlock(&GC_g_dl_ml);
    return(0);
}

# if defined(__STDC__) || defined(__cplusplus)
    int GC_unregister_disappearing_link(GC_PTR * link)
# else
    int GC_unregister_disappearing_link(link)
    GC_PTR * link;
# endif
{
    struct disappearing_link *curr_dl, *prev_dl;
    int index;
    
    PSM_mutex_lock_polling(&GC_g_dl_ml, SGC_polling);
    index = HASH2(link, GC_g_log_dl_table_size);
    if (((unsigned long)link & (ALIGNMENT-1))) goto out;
    prev_dl = 0; curr_dl = GC_g_dl_head[index];
    while (curr_dl != 0) {
        if (curr_dl -> dl_hidden_link == HIDE_POINTER(link)) {
            if (prev_dl == 0) {
                GC_g_dl_head[index] = dl_next(curr_dl);
            } else {
                dl_set_next(prev_dl, dl_next(curr_dl));
            }
            GC_g_dl_entries--;
	    PSM_mutex_unlock(&GC_g_dl_ml);
            GC_free((GC_PTR)curr_dl);
            return(1);
        }
        prev_dl = curr_dl;
        curr_dl = dl_next(curr_dl);
    }
out:
    PSM_mutex_unlock(&GC_g_dl_ml);
    return(0);
}

/* Possible finalization_marker procedures.  Note that mark stack	*/
/* overflow is handled by the caller, and is not a disaster.		*/
void GC_normal_finalize_mark_proc(p)
ptr_t p;
{
    DECLARE_MYENV
    DECLARE_MYENV3
    hdr * hhdr = HDR(p);
    word n_pushed; /* dummy by Endo */
    mse * limit = &(GC_my_mark_stack.buf[GC_my_mark_stack.size]);
    
    PUSH_OBJ((word *)p, hhdr, GC_my_mark_stack.top,
	     limit);
}

/* This only pays very partial attention to the mark descriptor.	*/
/* It does the right thing for normal and atomic objects, and treats	*/
/* most others as normal.						*/
void GC_ignore_self_finalize_mark_proc(p)
ptr_t p;
{
    DECLARE_MYENV3

    hdr * hhdr = HDR(p);
    word descr = hhdr -> hb_descr;
    ptr_t q, r;
    ptr_t scan_limit;
    ptr_t target_limit = p + WORDS_TO_BYTES(hhdr -> hb_sz) - 1;
    
    if ((descr & DS_TAGS) == DS_LENGTH) {
       scan_limit = p + descr - sizeof(word);
    } else {
       scan_limit = target_limit + 1 - sizeof(word);
    }
    for (q = p; q <= scan_limit; q += ALIGNMENT) {
    	r = *(ptr_t *)q;
    	if (r < p || r > target_limit) {
	    register ptr_t greatest_ha = GC_greatest_plausible_heap_addr;
	    register ptr_t least_ha = GC_least_plausible_heap_addr;
    	    GC_PUSH_ONE_HEAP((word)r, least_ha, greatest_ha);
    	}
    }
}

/*ARGSUSED*/
void GC_null_finalize_mark_proc(p)
ptr_t p;
{
}



/* Register a finalization function.  See gc.h for details.	*/
/* in the nonthreads case, we try to avoid disabling signals,	*/
/* since it can be expensive.  Threads packages typically	*/
/* make it cheaper.						*/
void GC_register_finalizer_inner(SGC_MYENV_ARG1 obj, fn, cd, ofn, ocd, mp)
SGC_MYENV_ARG_DECL2
GC_PTR obj;
GC_finalization_proc fn;
GC_PTR cd;
GC_finalization_proc * ofn;
GC_PTR * ocd;
finalization_mark_proc * mp;
{
    ptr_t base;
    struct finalizable_object * curr_fo, * prev_fo;
    int index;
    struct finalizable_object *new_fo;

    PSM_mutex_lock_polling(&GC_g_fo_ml, SGC_polling);
    if (GC_g_log_fo_table_size == -1
        || GC_g_fo_entries > ((word)1 << GC_g_log_fo_table_size)) {
    	GC_grow_table(SGC_MYENV_ARG1 (struct hash_chain_entry ***)(&GC_g_fo_head),
    		      &GC_g_log_fo_table_size);
#	ifdef PRINTSTATS
	    GC_printf1("Grew fo table to %lu entries\n",
	    		(unsigned long)(1 << GC_g_log_fo_table_size));
#	endif
    }
    /* in the THREADS case signals are disabled and we hold allocation	*/
    /* lock; otherwise neither is true.  Proceed carefully.		*/
    base = (ptr_t)obj;
    index = HASH2(base, GC_g_log_fo_table_size);
    prev_fo = 0; curr_fo = GC_g_fo_head[index];
    while (curr_fo != 0) {
        if (curr_fo -> fo_hidden_base == HIDE_POINTER(base)) {
            /* Interruption by a signal in the middle of this	*/
            /* should be safe.  The client may see only *ocd	*/
            /* updated, but we'll declare that to be his	*/
            /* problem.						*/
            if (ocd) *ocd = (GC_PTR) curr_fo -> fo_client_data;
            if (ofn) *ofn = curr_fo -> fo_fn;
            /* Delete the structure for base. */
                if (prev_fo == 0) {
                  GC_g_fo_head[index] = fo_next(curr_fo);
                } else {
                  fo_set_next(prev_fo, fo_next(curr_fo));
                }
            if (fn == 0) {
                GC_g_fo_entries--;
                  /* May not happen if we get a signal.  But a high	*/
                  /* estimate will only make the table larger than	*/
                  /* necessary.						*/
#		ifndef THREADS
                  GC_free((GC_PTR)curr_fo);
#		endif
            } else {
                curr_fo -> fo_fn = fn;
                curr_fo -> fo_client_data = (ptr_t)cd;
                curr_fo -> fo_mark_proc = mp;
		/* Reinsert it.  We deleted it first to maintain	*/
		/* consistency in the event of a signal.		*/
		if (prev_fo == 0) {
                  GC_g_fo_head[index] = curr_fo;
                } else {
                  fo_set_next(prev_fo, curr_fo);
                }
            }
	    PSM_mutex_unlock(&GC_g_fo_ml);
            return;
        }
        prev_fo = curr_fo;
        curr_fo = fo_next(curr_fo);
    }
    if (ofn) *ofn = 0;
    if (ocd) *ocd = 0;
    if (fn == 0) {
	PSM_mutex_unlock(&GC_g_fo_ml);
        return;
    }
      new_fo = (struct finalizable_object *)
    	GC_generic_malloc_inner(SGC_MYENV_ARG1 sizeof(struct finalizable_object),NORMAL);
    if (new_fo != 0) {
        new_fo -> fo_hidden_base = (word)HIDE_POINTER(base);
	new_fo -> fo_fn = fn;
	new_fo -> fo_client_data = (ptr_t)cd;
	new_fo -> fo_object_size = GC_size(base);
	new_fo -> fo_mark_proc = mp;
	fo_set_next(new_fo, GC_g_fo_head[index]);
	GC_g_fo_entries++;
	GC_g_fo_head[index] = new_fo;
    } else {
     	GC_g_finalization_failures++;
    }
    PSM_mutex_unlock(&GC_g_fo_ml);
}

# if defined(__STDC__)
    void GC_register_finalizer(void * obj,
			       GC_finalization_proc fn, void * cd,
			       GC_finalization_proc *ofn, void ** ocd)
# else
    void GC_register_finalizer(obj, fn, cd, ofn, ocd)
    GC_PTR obj;
    GC_finalization_proc fn;
    GC_PTR cd;
    GC_finalization_proc * ofn;
    GC_PTR * ocd;
# endif
{
    DECLARE_MYENV3
    GC_register_finalizer_inner(SGC_MYENV_ARG1 obj, fn, cd, ofn,
    				ocd, GC_normal_finalize_mark_proc);
}

# if defined(__STDC__)
    void GC_register_finalizer_ignore_self(void * obj,
			       GC_finalization_proc fn, void * cd,
			       GC_finalization_proc *ofn, void ** ocd)
# else
    void GC_register_finalizer_ignore_self(obj, fn, cd, ofn, ocd)
    GC_PTR obj;
    GC_finalization_proc fn;
    GC_PTR cd;
    GC_finalization_proc * ofn;
    GC_PTR * ocd;
# endif
{
    DECLARE_MYENV3
    GC_register_finalizer_inner(SGC_MYENV_ARG1 obj, fn, cd, ofn,
    				ocd, GC_ignore_self_finalize_mark_proc);
}

# if defined(__STDC__)
    void GC_register_finalizer_no_order(SGC_MYENV_ARG_DECL1 void * obj,
			       GC_finalization_proc fn, void * cd,
			       GC_finalization_proc *ofn, void ** ocd)
# else
    void GC_register_finalizer_no_order(SGC_MYENV_ARG1 obj, fn, cd, ofn, ocd)
SGC_MYENV_ARG_DECL2
    GC_PTR obj;
    GC_finalization_proc fn;
    GC_PTR cd;
    GC_finalization_proc * ofn;
    GC_PTR * ocd;
# endif
{
    GC_register_finalizer_inner(SGC_MYENV_ARG1 obj, fn, cd, ofn,
    				ocd, GC_null_finalize_mark_proc);
}


/* Called with world stopped.  Cause disappearing links to disappear,	*/
/* and invoke finalizers.						*/
void GC_finalize_gc(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    struct disappearing_link * curr_dl, * prev_dl, * next_dl;
    struct finalizable_object * curr_fo, * prev_fo, * next_fo;
    ptr_t real_ptr, real_link;
    register int i;
    int dl_size = (GC_g_log_dl_table_size == -1 ) ? 0 : (1 << GC_g_log_dl_table_size);
    int fo_size = (GC_g_log_fo_table_size == -1 ) ? 0 : (1 << GC_g_log_fo_table_size);
    
  /* Make disappearing links disappear */
    for (i = 0; i < dl_size; i++) {
      curr_dl = GC_g_dl_head[i];
      prev_dl = 0;
      while (curr_dl != 0) {
        real_ptr = (ptr_t)REVEAL_POINTER(curr_dl -> dl_hidden_obj);
        real_link = (ptr_t)REVEAL_POINTER(curr_dl -> dl_hidden_link);
        if (!GC_is_marked(real_ptr)) {
            *(word *)real_link = 0;
            next_dl = dl_next(curr_dl);
            if (prev_dl == 0) {
                GC_g_dl_head[i] = next_dl;
            } else {
                dl_set_next(prev_dl, next_dl);
            }
            GC_clear_mark_bit((ptr_t)curr_dl);
            GC_g_dl_entries--;
            curr_dl = next_dl;
        } else {
            prev_dl = curr_dl;
            curr_dl = dl_next(curr_dl);
        }
      }
    }
  /* Mark all objects reachable via chains of 1 or more pointers	*/
  /* from finalizable objects.						*/
#   ifdef PRINTSTATS
        if (GC_my_mark_state != MS_NONE) ABORT("Bad mark state");
#   endif
    for (i = 0; i < fo_size; i++) {
      for (curr_fo = GC_g_fo_head[i]; curr_fo != 0; curr_fo = fo_next(curr_fo)) {
        real_ptr = (ptr_t)REVEAL_POINTER(curr_fo -> fo_hidden_base);
        if (!GC_is_marked(real_ptr)) {
            (*(curr_fo -> fo_mark_proc))(real_ptr);
            while (!GC_mark_stack_empty(SGC_MYENV_ARG0)) GC_mark_from_mark_stack(SGC_MYENV_ARG0);
            if (GC_my_mark_state != MS_NONE) {
                /* Mark stack overflowed. Very unlikely. */
#		ifdef PRINTSTATS
		    if (GC_my_mark_state != MS_INVALID) ABORT("Bad mark state");
		    GC_printf0("Mark stack overflowed in finalization!!\n");
#		endif
		/* Make mark bits consistent again.  Forget about	*/
		/* finalizing this object for now.			*/
		    GC_set_mark_bit(real_ptr);
		    while (!GC_mark_some(SGC_MYENV_ARG0));
            }
            if (GC_is_marked(real_ptr)) {
                WARN("Finalization cycle involving %lx\n", real_ptr);
            }
        }
        
      }
    }
  /* Enqueue for finalization all objects that are still		*/
  /* unreachable.							*/
    GC_sum_app_stat.words_finalized = 0;
    for (i = 0; i < fo_size; i++) {
      curr_fo = GC_g_fo_head[i];
      prev_fo = 0;
      while (curr_fo != 0) {
        real_ptr = (ptr_t)REVEAL_POINTER(curr_fo -> fo_hidden_base);
        if (!GC_is_marked(real_ptr)) {
            GC_set_mark_bit(real_ptr);
            /* Delete from hash table */
              next_fo = fo_next(curr_fo);
              if (prev_fo == 0) {
                GC_g_fo_head[i] = next_fo;
              } else {
                fo_set_next(prev_fo, next_fo);
              }
              GC_g_fo_entries--;
            /* Add to list of objects awaiting finalization.	*/
              fo_set_next(curr_fo, GC_g_finalize_now);
              GC_g_finalize_now = curr_fo;
              /* unhide object pointer so any future collections will	*/
              /* see it.						*/
              curr_fo -> fo_hidden_base = 
              		(word) REVEAL_POINTER(curr_fo -> fo_hidden_base);
	      GC_sum_app_stat.words_finalized +=
                 	ALIGNED_WORDS(curr_fo -> fo_object_size)
              		+ ALIGNED_WORDS(sizeof(struct finalizable_object));
#	    ifdef PRINTSTATS
              if (!GC_is_marked((ptr_t)curr_fo)) {
                ABORT("GC_finalize: found accessible unmarked object\n");
              }
#	    endif
            curr_fo = next_fo;
        } else {
            prev_fo = curr_fo;
            curr_fo = fo_next(curr_fo);
        }
      }
    }
  /* Remove dangling disappearing links. */
    for (i = 0; i < dl_size; i++) {
      curr_dl = GC_g_dl_head[i];
      prev_dl = 0;
      while (curr_dl != 0) {
        real_link = GC_base((ptr_t)REVEAL_POINTER(curr_dl -> dl_hidden_link));
        if (real_link != 0 && !GC_is_marked(real_link)) {
            next_dl = dl_next(curr_dl);
            if (prev_dl == 0) {
                GC_g_dl_head[i] = next_dl;
            } else {
                dl_set_next(prev_dl, next_dl);
            }
            GC_clear_mark_bit((ptr_t)curr_dl);
            GC_g_dl_entries--;
            curr_dl = next_dl;
        } else {
            prev_dl = curr_dl;
            curr_dl = dl_next(curr_dl);
        }
      }
    }
}

/* Invoke finalizers for all objects that are ready to be finalized.	*/
/* Should be called without allocation lock.				*/
int GC_invoke_finalizers()
{
    register struct finalizable_object * curr_fo;
    register int count = 0;
    
    while (GC_g_finalize_now != 0) {
	PSM_mutex_lock_polling(&GC_g_fo_ml, SGC_polling);
    	curr_fo = GC_g_finalize_now;
	if (curr_fo != 0) GC_g_finalize_now = fo_next(curr_fo);
	PSM_mutex_unlock(&GC_g_fo_ml);
	if (curr_fo == 0) break;
 	fo_set_next(curr_fo, 0);
    	(*(curr_fo -> fo_fn))((ptr_t)(curr_fo -> fo_hidden_base),
    			      curr_fo -> fo_client_data);
    	curr_fo -> fo_client_data = 0;
	++count;
#	ifdef UNDEFINED
	    /* This is probably a bad idea.  It throws off accounting if */
	    /* nearly all objects are finalizable.  O.w. it shouldn't	 */
	    /* matter.							 */
    	    GC_free((GC_PTR)curr_fo);
#	endif
    }
    return count;
}

# ifdef __STDC__
    GC_PTR GC_call_with_alloc_lock(GC_fn_type fn,
    					 GC_PTR client_data)
# else
    GC_PTR GC_call_with_alloc_lock(fn, client_data)
    GC_fn_type fn;
    GC_PTR client_data;
# endif
{
    GC_PTR result;
    
    PSM_mutex_lock_polling(&GC_g_alloc_ml, SGC_polling);
    result = (*fn)(client_data);
    PSM_mutex_unlock(&GC_g_alloc_ml);
    return(result);
}

