/*
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */
/* Endo, January 13, 1999 */

#ifndef _BLOCKSYNC_H
#define _BLOCKSYNC_H

#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD

#include <sys/ipc.h>
#include <sys/sem.h>

typedef struct sh_semgroup {
    struct sh_semgroup *next;
    int semid;
    char *used_sem_num;
} sh_semgroup_t;

#endif

#ifdef SDSM_TEST
#include <sdsm.h>
#endif

/* Blocking sync manager structure */
typedef struct {
#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD
    int nsem_per_group;
    sh_semgroup_t *semgroup_list;
    int nsem;
    int ngroup;
#else
    int dummy; /* some compiler doesn't allow empty structures. */
#endif
} sh_blocksync_t;

/* Semaphore */
typedef struct {
    sh_blocksync_t *bsp;
#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD
    sh_semgroup_t *sgp;
    short sem_num;
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    mutex_t st_ml;
#endif
#if defined SGC_PTHREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    pthread_mutex_t pt_ml;
#endif
#ifdef SDSM_TEST
    S_mutex_t sdsm_ml;
#endif
} sh_bssem_t;

/* Barrier */
typedef struct {
    sh_blocksync_t *bsp;
    int np;
    int count;
    int sense;
#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD
    sh_bssem_t esems[2]; /* Entering barrier */
    sh_bssem_t lsems[2]; /* Leaving barrier */
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    mutex_t st_ml;
    cond_t st_cv;
#endif
#if defined SGC_PTHREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    pthread_mutex_t pt_ml;
    pthread_cond_t pt_cv;
#endif
} sh_bsbar_t;

/* Prototype */

int GC_blocksync_init(sh_blocksync_t *bsp, int maxsem, int maxbar);
void GC_blocksync_end(sh_blocksync_t *bsp);
int GC_bssem_init(sh_bssem_t *semp, sh_blocksync_t *bsp);
void GC_bssem_end(sh_bssem_t *semp);
void GC_bssem_clear(sh_bssem_t *semp);
void GC_bssem_signal(sh_bssem_t *semp);
void GC_bssem_wait(sh_bssem_t *semp);
int GC_bsbar_init(sh_bsbar_t *barp, sh_blocksync_t *bsp, int np);
void GC_bsbar_sync(sh_bsbar_t *barp, int isleader);
void GC_bsbar_end(sh_bsbar_t *barp);

#endif /* _BLOCKSYNC_H */

