/*
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */
/* Endo, August 17, 1999 */

#ifndef SHAREDLOG_H
#define SHAREDLOG_H

/* LDLOG: log about time-breakdown in each collection. */

#define GC_LDC_S 0
#define GC_LDC_E 1

#define GC_LDK_MFMS 0
#define GC_LDK_SEARCH 1
#define GC_LDK_PUSHL 2
#define GC_LDK_DTERM 3
#define GC_LDK_WMB 4
#define GC_LDK_WMQ 5
#define GC_LDK_SLONG 6
#define GC_LDK_PROOT 7
#define GC_LDK_SROOT 8
#define GC_LDK_WHFL 9
#define GC_LDK_WRL 10
#define GC_LDK_WREQB 11

#ifdef DO_LDLOG
static char *GC_const_ldlog_comname[] = {
    "S",
    "E",
};

static char *GC_const_ldlog_kindname[] = {
    "MFMS",
    "SEARCH",
    "PUSHL",
    "DTERM",
    "WMB",
    "WMQ",
    "SLONG",
    "PROOT",
    "SROOT",
    "WHFL",
    "WRL",
    "WREQB",
};

#  define GC_LDLOG_INIT() GC_ldlog_init()
#  define GC_LDLOG_INIT_LOCAL(envp) GC_ldlog_init_local(envp)
#  define GC_LDLOG_INIT_PHASE() GC_ldlog_init_phase()
#  define GC_LDLOG_END_PHASE() GC_ldlog_end_phase()
#  define GC_LDLOG_PUT(com, kind) GC_ldlog_put(SGC_MYENV_ARG1 com, kind)

#define LDLOG_NUM 60000

struct lg_ldlog {
    long index;
    struct ldlog_entry {
	char com;
	char kind;
	struct timeval tv;
    } data[LDLOG_NUM];
};

struct g_ldlog {
    long phase;
    long doing;
    struct timeval st;
};

#else
#  define GC_LDLOG_INIT()
#  define GC_LDLOG_INIT_LOCAL(pno)
#  define GC_LDLOG_INIT_PHASE()
#  define GC_LDLOG_END_PHASE()
#  define GC_LDLOG_PUT(com, kind)
#endif


/* ALLOCLOG: log about the frequency of GC_malloc by user. */

#define GC_ALLOCC_S 0
#define GC_ALLOCC_E 1

#ifdef DO_ALLOCLOG
#  define GC_ALLOCLOG_INIT() GC_alloclog_init()
#  define GC_ALLOCLOG_INIT_LOCAL(envp) GC_alloclog_init_local(envp)
#  define GC_ALLOCLOG_PUT(com) GC_alloclog_put(SGC_MYENV_ARG1 com)
#  define GC_ALLOCLOG_END_LOCAL(envp) GC_alloclog_end_local(envp)
#  define GC_ALLOCLOG_END() GC_alloclog_end()

#define ALLOCLOG_NUM 30000

struct la_alloclog {
    long index;
    struct alloclog_entry {
	int com;
	struct timeval tv;
    } data[ALLOCLOG_NUM];
};

struct g_alloclog {
    PSM_mutex_t ml;
    int pcount;
    char fn[64];
    FILE *fp;
    struct timeval st;
};

static char *GC_const_alloclog_comname[] = {
    "S",
    "E",
};

#else
#  define GC_ALLOCLOG_INIT()
#  define GC_ALLOCLOG_INIT_LOCAL(envp)
#  define GC_ALLOCLOG_PUT(com)
#  define GC_ALLOCLOG_END_LOCAL(envp)
#  define GC_ALLOCLOG_END()
#endif


/* MARKLOG: log about GC-speed = (GC-workload/GC-time) */

#define GC_LOGKIND_N_MARKED 0 /* Mark */
#define GC_LOGKIND_N_HASMARKED 1 /* Mark */
#define GC_LOGKIND_N_NONPTR 2 /* Mark */
#define GC_LOGKIND_N_FREED 3 /* Sweep */
#define GC_LOGKIND_N_TORECLAIM 4 /* Sweep */
#define GC_LOGKIND_N_MOVED 5 /* Mark ld */
#define GC_LOGKIND_N_NOHDR 6 /* Sweep */

#ifdef DO_MARKLOG
#  define GC_MARKLOG_INIT() GC_marklog_init()
#  define GC_MARKLOG_INIT_LOCAL(envp) GC_marklog_init_local(envp)
#  define GC_MARKLOG_INIT_PHASE() GC_marklog_init_phase()
#  define GC_MARKLOG_END_PHASE() GC_marklog_end_phase()
#  define GC_MARKLOG_START_CLOCK() GC_marklog_start_clock()
#  define GC_MARKLOG_STOP_CLOCK() GC_marklog_stop_clock()
#  define GC_FAST_MARKLOG_PUT(k, c) \
{GC_my_marklogp->count[k] += c;}
#  define GC_MARKLOG_PUT(k, c) GC_marklog_put(SGC_MYENV_ARG1 k, c)

#define MARKLOG_KIND 8

struct l_marklog {
    long count[MARKLOG_KIND];
};

#define MARKLOG_CLOCK 3

struct g_marklog {
    struct timeval rsts[MARKLOG_CLOCK];
    struct timeval rets[MARKLOG_CLOCK];
    int currentclock;
};

#else
#  define GC_MARKLOG_INIT()
#  define GC_MARKLOG_INIT_LOCAL(envp)
#  define GC_MARKLOG_INIT_PHASE()
#  define GC_MARKLOG_END_PHASE()
#  define GC_MARKLOG_START_CLOCK()
#  define GC_MARKLOG_STOP_CLOCK()
#  define GC_FAST_MARKLOG_PUT(k, c)
#  define GC_MARKLOG_PUT(k, c)
#endif


/* PICLOG: log with performance counter. (currently sparc/mips only) */

#define GC_PICK_ETC 0
#define GC_PICK_SOBJ 2 /* scanning objects in mark phase */
#define GC_PICK_MMB 4 /* writing mark bits in mark phase */

#ifdef DO_PICLOG
#  define GC_PICLOG_INIT() GC_piclog_init()
#  define GC_PICLOG_INIT_LOCAL(envp) GC_piclog_init_local(envp)
#  define GC_PICLOG_INIT_PHASE_LOCAL() GC_piclog_init_phase_local()
#  define GC_PICLOG_CHANGE_MODE(n) GC_piclog_change_mode(n)
#  define GC_PICLOG_ADD(n, c0, c1) GC_piclog_add(SGC_MYENV_ARG1 n, c0, c1)
#  define GC_PICLOG_END_PHASE_LOCAL() GC_piclog_end_phase_local()
#  define GC_PICLOG_END_PHASE() GC_piclog_end_phase()

#define PICLOG_MODE 4

struct l_piclog {
    int mode;
    long start[2];
    long count[PICLOG_MODE][2];
    PERFC_handler_t handler;
};

struct g_piclog {
    int dummy0; /* some compilers inhibit empty structures */
};

#else
#  define GC_PICLOG_INIT()
#  define GC_PICLOG_INIT_LOCAL(envp)
#  define GC_PICLOG_INIT_PHASE_LOCAL()
#  define GC_PICLOG_CHANGE_MODE(n)
#  define GC_PICLOG_ADD(n, c0, c1)
#  define GC_PICLOG_PEADD(pe, n, c0, c1)
#  define GC_PICLOG_END_PHASE_LOCAL()
#  define GC_PICLOG_END_PHASE()

#  define GC_PICLOG_MEMBAR() 
#  define GC_PICLOG_RDPIC01(n0, n1)
#endif


/* GCRLOG: log about rate of GC-time to application-time. require MARKLOG */

#ifdef DO_GCRLOG
struct g_gcrlog {
    struct timeval st;
    struct timeval gc_st;

    /* total */
    long gc_us;
    long sweep_us;
    long mark_us;

    struct timeval clocks[10];
};

#  define GC_GCRLOG_INIT() GC_gcrlog_init()
#  define GC_GCRLOG_INIT_GC() GC_gcrlog_init_gc()
#  define GC_GCRLOG_SET_CLOCK(i) GC_gcrlog_set_clock(i)
#  define GC_GCRLOG_END_GC() GC_gcrlog_end_gc()
#  define GC_GCRLOG_END() GC_gcrlog_end()
#else
#  define GC_GCRLOG_INIT()
#  define GC_GCRLOG_INIT_GC()
#  define GC_GCRLOG_SET_CLOCK(i)
#  define GC_GCRLOG_END_GC()
#  define GC_GCRLOG_END()
#endif


/* MSLOG: log about mark stack usage. */

#define GC_MSC_PUSH 0
#define GC_MSC_POP 1
#define GC_MSC_STEAL 2
#define GC_MSC_PUSH_POP 3 /* PUSH and immediate POP */
#define GC_MSC_STEAL_POP 4 /* STEAL and immediate POP */

#ifdef DO_MSLOG
static char *GC_const_mslog_comname[] = {
    "PUSH",
    "POP",
    "STEAL",
    "PUSH_POP",
    "STEAL_POP",
};

#  define GC_MSLOG_INIT() GC_mslog_init()
#  define GC_MSLOG_INIT_LOCAL(envp) GC_mslog_init_local(envp)
#  define GC_MSLOG_INIT_PHASE() GC_mslog_init_phase()
#  define GC_MSLOG_END_PHASE() GC_mslog_end_phase()
#  define GC_MSLOG_PUT(com, obj) GC_mslog_put(SGC_MYENV_ARG1 com, obj)

#define MSLOG_NUM 300000

struct lg_mslog {
    long index;
    long time;
    struct mslog_entry {
	long timestamp;
	char com;
	word *obj;
    } data[MSLOG_NUM];
};

struct g_mslog {
    long phase;
    long doing;
};

#else
#  define GC_MSLOG_INIT()
#  define GC_MSLOG_INIT_LOCAL(envp)
#  define GC_MSLOG_INIT_PHASE()
#  define GC_MSLOG_END_PHASE()
#  define GC_MSLOG_PUT(com, obj)
#endif

/* CACHELOG: log about approximate cache misses */

#ifdef DO_CACHELOG
struct l_cachelog {
    hdr *prevhdr;
    int mbhit;
    int mbmiss;
    int objmiss;
};

#define GC_CACHELOG_INIT_LOCAL(envp) GC_cachelog_init_local(envp)
#define GC_CACHELOG_INIT_PHASE() GC_cachelog_init_phase()
#define GC_CACHELOG_BEFORE_LOOP() GC_cachelog_before_loop(SGC_MYENV_ARG0)
#define GC_CACHELOG_PUT_MBHDR(hhdr) GC_cachelog_put_mbhdr(SGC_MYENV_ARG1 hhdr)
#define GC_CACHELOG_PUT_OBJSIZE(s) GC_cachelog_put_objsize(SGC_MYENV_ARG1 s)
#define GC_CACHELOG_END_PHASE() GC_cachelog_end_phase()

#else
#define GC_CACHELOG_INIT_LOCAL(envp)
#define GC_CACHELOG_INIT_PHASE()
#define GC_CACHELOG_BEFORE_LOOP()
#define GC_CACHELOG_PUT_MBHDR(hdr)
#define GC_CACHELOG_PUT_OBJSIZE(s)
#define GC_CACHELOG_END_PHASE()
#endif

#endif /* SHAREDLOG_H */
