/*
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */
/* Endo, June 3, 1999 */

#include <stdio.h>
#include "gc_priv.h"

static int GC_detect_termination GC_PROTO((SGC_MYENV_ARG_DECL0)); 

/*******************************************************************/
#if DO_LOAD_DIST == 3

#define DO_PUSH_ALMOST 1
#define DO_MS_PUSH_LOCK 1

/* don't use GC_my_envp in this function. */
int GC_ld_init_local(GC_thread envp)
{
    /* Allocate stealable mark queue. */
    GC_lg_smq(envp).size = 8*HBLKSIZE;
    GC_lg_smq(envp).buf = 
	(char *)GET_MEM(GC_lg_smq(envp).size*
			sizeof(struct ms_entry));
    if (GC_lg_smq(envp).buf == NULL) {
	fprintf(stderr, "[GC_ld_init_local] Cannot alloc stealable_mark_queue.\n");
    }
    GC_touch_memory_region(GC_lg_smq(envp).buf,
			   GC_lg_smq(envp).size*sizeof(struct ms_entry));
    return 0;
}

int GC_ld_init_phase()
{
    GC_thread envp;

#if DO_OPT_NSB == 1
    PSM_mutex_init(&GC_g_ld.dterm_ml);
    GC_g_ld.all_finished = 0;
#else
    GC_g_ld.busy_count = GC_count_gcproc_envps();
    PSM_mutex_init(&GC_g_ld.busy_count_ml);
    GC_g_ld.nonempty_count = 0;
    PSM_mutex_init(&GC_g_ld.nonempty_count_ml);
#endif

    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	GC_lg_smq(envp).in = GC_lg_smq(envp).buf;
	GC_lg_smq(envp).out = GC_lg_smq(envp).buf;
	PSM_mutex_init(&GC_lg_smq(envp).ml);
#if DO_OPT_NSB == 1
	GC_lg_ld(envp).busy_flag = 1;
	GC_lg_ld(envp).nonempty_flag = 0;
#endif
    }
    return 0;
}

int GC_push_overload(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    register mse *push_limit;
    register mse *p;
    register mse *p2;
    int i;
    int old_ne;
    int new_ne;
    int n;
    int procno;
    register hdr * hhdr; 
    int smq_size = GC_myg_smq.size;
    register mse *smq_in = (mse *)GC_myg_smq.in;
    mse *smq_out = (mse *)GC_myg_smq.out;
    mse *smq_buf = (mse *)GC_myg_smq.buf;
    register mse *smq_limit = smq_buf+smq_size;
    
    old_ne = (smq_in >= smq_out)? (smq_in-smq_out): (smq_in+smq_size-smq_out);
    if (old_ne) {
	/* There are some objects in e.m.q */
	return 0;
    }
    
    n = GC_my_mark_stack.top - GC_my_mark_stack.bottom + 1;
#if DO_PUSH_ALMOST == 1
    n = n-1; /* leave 1 entry */
#else
    n = n/4;
#endif
    if (n <= 0) {
	/* No moveable objects in mark stack */
	return 0;
    }
    
#if DO_MS_PUSH_LOCK == 1
    MUTEX_LOCK_LOG(&GC_myg_smq.ml, GC_LDK_WMQ); 
#endif

    GC_LDLOG_PUT(GC_LDC_S, GC_LDK_PUSHL);
    GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_MOVED, 1);
    
    if (old_ne + n >= smq_size) {
	fprintf(stderr, "!!! stelable mark queue overflow (%08x + %x >= %08x)!!!\n", old_ne, n, smq_size);
	exit(1);
    }
    
    push_limit = GC_my_mark_stack.top-n+1;

    for (p = GC_my_mark_stack.top; p >= push_limit; p--) {
	COPY_MSE(smq_in, p);
	smq_in ++;
	if (smq_in >= smq_limit) {
	    smq_in = smq_buf;
	}
    }
    GC_my_mark_stack.top = push_limit-1;
    GC_myg_smq.in = (char *)smq_in;
    
    new_ne = (smq_in >= smq_out)? (smq_in-smq_out): (smq_in+smq_size-smq_out);
    if (old_ne == 0 && new_ne > 0) {
#if DO_OPT_NSB == 1
	GC_myg_ld.nonempty_flag = 1;
	GC_g_ld.flag_modified = 1;
#else
	PSM_counter_add(&GC_g_ld.nonempty_count, 1);
#endif
    }
    
#if DO_MS_PUSH_LOCK == 1
    PSM_mutex_unlock(&GC_myg_smq.ml);
#endif
    GC_LDLOG_PUT(GC_LDC_E, GC_LDK_PUSHL);
    return 0;
}

static int GC_detect_termination(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
#if DO_OPT_NSB == 1
    /* for Non-serializing barrier */
    if (GC_g_ld.all_finished) {
	return 1;
    }
    if (PSM_mutex_trylock(&GC_g_ld.dterm_ml)) {
	GC_thread envp2;
	GC_LDLOG_PUT(GC_LDC_S, GC_LDK_DTERM);
	GC_g_ld.flag_modified = 0;
	
	for (envp2 = GC_g_gcproc_envps.head; envp2; envp2 = envp2->next) {
	    if (GC_l_initialized(envp2) == 0) continue;
	    if (GC_l_terminated(envp2) != 0) continue;
	    if (GC_lg_ld(envp2).busy_flag ||
		GC_lg_ld(envp2).nonempty_flag) {
		break;
	    }
	}
	GC_LDLOG_PUT(GC_LDC_E, GC_LDK_DTERM);
	if (envp2 == NULL &&
	    GC_g_ld.flag_modified == 0) {
	    /* All PE and queues are idle! */
	    PSM_mutex_unlock(&GC_g_ld.dterm_ml);
	    GC_g_ld.all_finished = 1;
	    return 1;
	}
	PSM_mutex_unlock(&GC_g_ld.dterm_ml);
    }
#else
    {
	int volatile *pbc;
	int volatile *pnc;
	pbc = &GC_g_ld.busy_count;
	pnc = &GC_g_ld.nonempty_count;
	if (*pbc == 0 && *pnc == 0) {
	    /* All processes idle, so Terminate */
	    return 1;
	}
    }
#endif
    return 0;
}

int GC_get_work_or_terminate(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    register int n;
    register mse *p;
    int endflag = 0;
    GC_thread envp;
    int i;
    int nloop = 0;
    int old_ne;
    int new_ne;
    int smq_size;
    mse *smq_in;
    register mse *smq_out;
    mse *smq_buf;
    register mse *smq_limit;

#if DO_OPT_NSB == 1
    GC_myg_ld.busy_flag = 0;
#else
    PSM_counter_add(&GC_g_ld.busy_count, -1);
    if (GC_g_ld.busy_count == 0 && GC_g_ld.nonempty_count == 0) {
	endflag = 1;
	return endflag;
    }
#endif

    GC_LDLOG_PUT(GC_LDC_S, GC_LDK_SEARCH);
    while (1) {
#ifdef DO_REQ_REMOTE
	if (GC_myg_reqbuf.validsize > 0) {
#if DO_OPT_NSB == 1
	    GC_myg_ld.busy_flag = 1;
	    GC_g_ld.flag_modified = 1;
#else
	    PSM_counter_add(&GC_g_ld.busy_count, 1);
#endif
	    endflag = 0;
	    break;
	}
#endif /* DO_REQ_REMOTE */
	/* check my s.m.q. and others. */
	for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	    if (GC_l_initialized(envp) == 0) continue;
	    if (GC_l_terminated(envp) != 0) continue;

	    if (nloop > 0 && GC_detect_termination(SGC_MYENV_ARG0)) {
		/* mark phase has finished! */
		endflag = 1;
		break;
	    }

#if DO_OPT_SLQ == 1
	    /* This may improve progress. */
	    if (PSM_mutex_trylock(&GC_lg_smq(envp).ml) == 0) {
		continue;
	    }
#else
	    MUTEX_LOCK_LOG(&GC_lg_smq(envp).ml, GC_LDK_WMQ);
#endif
	    smq_size = GC_lg_smq(envp).size;
	    smq_in = (mse *)GC_lg_smq(envp).in;
	    smq_out = (mse *)GC_lg_smq(envp).out;
	    smq_buf = (mse *)GC_lg_smq(envp).buf;
	    smq_limit = smq_buf + smq_size;

	    old_ne = (smq_in >= smq_out)? (smq_in-smq_out):
		(smq_in+smq_size-smq_out);

	    if (old_ne <= 0) {
		PSM_mutex_unlock(&GC_lg_smq(envp).ml);
		continue;
	    }

	    /* Steal any work */
	    n = (old_ne + 1) / 2;
	    GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_MOVED, 1);
	    p = GC_my_mark_stack.top+1;
	    /* Scan half(all) entries in queue. (ascending order) */
	    for (i = 0; i < n; i++) {
		COPY_MSE(p, smq_out);
		GC_MSLOG_PUT(GC_MSC_STEAL, p->mse_start);
		smq_out++;
		p ++;
		if (smq_out >= smq_limit) {
		    smq_out = smq_buf;
		}
	    }
	    GC_my_mark_stack.top = p-1;
	    GC_lg_smq(envp).out = (char *)smq_out;
#if DO_OPT_NSB == 1
	    GC_myg_ld.busy_flag = 1;
	    GC_g_ld.flag_modified = 1;
#else
	    PSM_counter_add(&GC_g_ld.busy_count, 1);
#endif
	    new_ne = (smq_in >= smq_out)? (smq_in-smq_out):
		(smq_in+smq_size-smq_out);
	    if (old_ne > 0 && new_ne == 0) {
		/* This queue became empty. */
#if DO_OPT_NSB == 1
		GC_lg_ld(envp).nonempty_flag = 0;
#else
		PSM_counter_add(&GC_g_ld.nonempty_count, -1);
#endif
	    }
	    PSM_mutex_unlock(&GC_lg_smq(envp).ml);
	    endflag = 0;
	    break;
	}
	if (envp != NULL) break;
	nloop++;
    }

    GC_LDLOG_PUT(GC_LDC_E, GC_LDK_SEARCH);
    return endflag;
}

#endif /* DO_LOAD_DIST == 3 */

/*******************************************************************/
#if DO_LOAD_DIST == 4
#define DO_OPT_SEQ 0 /* Skip empty queue without lock */

/* don't use GC_my_envp in this function. */
int GC_ld_init_local(GC_thread envp)
{
    return 0;
}

/*******************************************************************/
int GC_ld_init_phase()
{
    GC_thread envp;
#if DO_OPT_NSB == 1
    PSM_mutex_init(&GC_g_ld.dterm_ml);
    GC_g_ld.all_finished = 0;
#else
    GC_g_ld.busy_count = GC_count_gcproc_envps();
    PSM_mutex_init(&GC_g_ld.busy_count_ml);
#endif

#if DO_OPT_NSB == 1
    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	GC_lg_ld(envp).busy_flag = 1;
    }
#endif
    return 0;
}

/*******************************************************************/
static int GC_detect_termination(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
#if DO_OPT_NSB == 1
    /* for Non-serializing barrier */
    if (GC_g_ld.all_finished) {
	return 1;
    }
    if (PSM_mutex_trylock(&GC_g_ld.dterm_ml)) {
	GC_thread envp2;
	GC_LDLOG_PUT(GC_LDC_S, GC_LDK_DTERM);
	GC_g_ld.flag_modified = 0;
	
	for (envp2 = GC_g_gcproc_envps.head; envp2; envp2 = envp2->next) {
	    if (GC_l_initialized(envp2) == 0) continue;
	    if (GC_l_terminated(envp2) != 0) continue;
	    if (GC_lg_ld(envp2).busy_flag) {
		break;
	    }
	}
	GC_LDLOG_PUT(GC_LDC_E, GC_LDK_DTERM);
	if (envp2 == NULL &&
	    GC_g_ld.flag_modified == 0) {
	    /* All PE and queues are idle! */
	    PSM_mutex_unlock(&GC_g_ld.dterm_ml);
	    GC_g_ld.all_finished = 1;
	    return 1;
	}
	PSM_mutex_unlock(&GC_g_ld.dterm_ml);
    }
#else
    {
	int volatile *pbc;
	int volatile *pnc;
	pbc = &GC_g_ld.busy_count;
	if (*pbc == 0) {
	    /* All processes idle, so Terminate */
	    return 1;
	}
    }
#endif
    return 0;
}

int GC_push_overload(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    /* do nothing */
    return 0;
}

int GC_get_work_or_terminate(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    int endflag = 0;
    GC_thread envp;
    int nloop = 0;
    mark_stack_t *from_ms;
    register mse *from_ms_bottom;
    register mse *from_ms_top;
    mark_stack_t *to_ms;
    int old_ne;

#if DO_OPT_NSB == 1
    GC_myg_ld.busy_flag = 0;
#else
    PSM_counter_add(&GC_g_ld.busy_count, -1);
    if (GC_g_ld.busy_count == 0) {
	endflag = 1;
	return endflag;
    }
#endif

    GC_LDLOG_PUT(GC_LDC_S, GC_LDK_SEARCH);
    to_ms = &GC_my_mark_stack;
    while (1) {
#ifdef DO_REQ_REMOTE
	if (GC_my_reqbuf.validsize > 0) {
#if DO_OPT_NSB == 1
	    GC_myg_ld.busy_flag = 1;
	    GC_g_ld.flag_modified = 1;
#else
	    PSM_counter_add(&GC_g_ld.busy_count, 1);
#endif
	    endflag = 0;
	    break;
	}
#endif
	/* See other mark stack. */
	for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	    if (GC_l_initialized(envp) == 0) continue;
	    if (GC_l_terminated(envp) != 0) continue;

	    /* Termination Detection */
	    if (nloop > 0 && GC_detect_termination(SGC_MYENV_ARG0)) {
		/* mark phase has finished! */
		endflag = 1;
		break;
	    }

	    if (envp == GC_my_envp) {
		/* Skip self */
		continue;
	    }

	    from_ms = &GC_l_mark_stack(envp);
#if DO_OPT_SEQ == 1
	    /* Skip empty queue(stack). */
	    old_ne = from_ms->top - from_ms->bottom + 1;
	    if (old_ne <= 0) {
		continue;
	    }
#endif
#if DO_OPT_SLQ == 1
	    /* This may improve progress. */
	    if (PSM_mutex_trylock(&from_ms->ml) == 0) {
		continue;
	    }
#else
	    MUTEX_LOCK_LOG(&from_ms->ml, GC_LDK_WMQ);
#endif
	    from_ms_top = from_ms->top;
	    from_ms_bottom = from_ms->bottom;
	    /* from_ms->top may be incremented by owner without lock, */
	    /* but that is harmless. */
	    old_ne = from_ms_top - from_ms_bottom + 1;

	    if (old_ne <= 0) {
		PSM_mutex_unlock(&from_ms->ml);
		continue;
	    }
	    /* Found work! */
	    PSM_mutex_lock(&to_ms->ml);
	    GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_MOVED, 1);
	    /* Pushing onto to_ms without lock is OK */
#if DO_OPT_SLO == 2
	    {
		int objsz = (from_ms->bottom->mse_descr) & ~DS_TAGS;
		ptr_t p = from_ms->bottom->mse_start;
		if (objsz >= CUTBYTES_TO_PUSH &&
		    (from_ms->bottom->mse_descr & DS_TAGS) == DS_LENGTH) {
		    /* steal the second half of objects */
		    int remainsz = WORDS_TO_BYTES(BYTES_TO_WORDS(objsz/2));
		    COPY_MSE(to_ms->top+1, from_ms_bottom);
		    (to_ms->top+1)->mse_start = (word *)(p + remainsz);
		    (to_ms->top+1)->mse_descr =
			(((objsz-remainsz+DS_TAGS) & ~DS_TAGS) | DS_LENGTH);
		    from_ms->bottom->mse_start = (word *)p;
		    from_ms->bottom->mse_descr =
			(((remainsz+DS_TAGS) & ~DS_TAGS) | DS_LENGTH);
		    to_ms->top++;
		    GC_MSLOG_PUT(GC_MSC_STEAL, (word *)(p + remainsz));
		}
		else {
		    COPY_MSE(to_ms->top+1, from_ms_bottom);
		    GC_MSLOG_PUT(GC_MSC_STEAL, (to_ms->top+1)->mse_start);
		    to_ms->top++;
		    from_ms->bottom++;
		}
	    }
#else
	    COPY_MSE(to_ms->top+1, from_ms_bottom);
	    GC_MSLOG_PUT(GC_MSC_STEAL, (to_ms->top+1)->mse_start);
	    to_ms->top++;
	    from_ms->bottom++;
#endif

#if DO_OPT_NSB == 1
	    GC_myg_ld.busy_flag = 1;
	    GC_g_ld.flag_modified = 1;
#else
	    PSM_counter_add(&GC_g_ld.busy_count, 1);
#endif
	    PSM_mutex_unlock(&to_ms->ml);
	    PSM_mutex_unlock(&from_ms->ml);
	    endflag = 0;
	    break;
	}
	if (envp != NULL) break;
	nloop++;
    }

    GC_LDLOG_PUT(GC_LDC_E, GC_LDK_SEARCH);

    return endflag;
}

#endif /* DO_LOAD_DIST == 4 */

/*******************************************************************/
#if DO_LOAD_DIST == 5
#define DO_OPT_SEQ 0 /* Skip empty queue without lock */

/* don't use GC_my_envp in this function. */
int GC_ld_init_local(GC_thread envp)
{
    return 0;
}

/*******************************************************************/
int GC_ld_init_phase()
{
    GC_thread envp;
#if DO_OPT_NSB == 1
    PSM_mutex_init(&GC_g_ld.dterm_ml);
    GC_g_ld.all_finished = 0;
#else
    GC_g_ld.busy_count = GC_count_gcproc_envps();
    PSM_mutex_init(&GC_g_ld.busy_count_ml);
#endif

#if DO_OPT_NSB == 1
    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	GC_lg_ld(envp).busy_flag = 1;
    }
#endif
    return 0;
}

/*******************************************************************/
static int GC_detect_termination(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
#if DO_OPT_NSB == 1
    /* for Non-serializing barrier */
    if (GC_g_ld.all_finished) {
	return 1;
    }
    if (PSM_mutex_trylock(&GC_g_ld.dterm_ml)) {
	GC_thread envp2;
	GC_LDLOG_PUT(GC_LDC_S, GC_LDK_DTERM);
	GC_g_ld.flag_modified = 0;
	
	for (envp2 = GC_g_gcproc_envps.head; envp2; envp2 = envp2->next) {
	    if (GC_l_initialized(envp2) == 0) continue;
	    if (GC_l_terminated(envp2) != 0) continue;
	    if (GC_lg_ld(envp2).busy_flag) {
		break;
	    }
	}
	GC_LDLOG_PUT(GC_LDC_E, GC_LDK_DTERM);
	if (envp2 == NULL &&
	    GC_g_ld.flag_modified == 0) {
	    /* All PE and queues are idle! */
	    PSM_mutex_unlock(&GC_g_ld.dterm_ml);
	    GC_g_ld.all_finished = 1;
	    return 1;
	}
	PSM_mutex_unlock(&GC_g_ld.dterm_ml);
    }
#else
    {
	int volatile *pbc;
	int volatile *pnc;
	pbc = &GC_g_ld.busy_count;
	if (*pbc == 0) {
	    /* All processes idle, so Terminate */
	    return 1;
	}
    }
#endif
    return 0;
}

int GC_push_overload(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    int nmse;

    MUTEX_LOCK_LOG(&GC_my_mark_stack.ml, GC_LDK_WMQ);
    if (GC_my_mark_stack.watermark - GC_my_mark_stack.bottom > 0) {
	PSM_mutex_unlock(&GC_my_mark_stack.ml);
	return 0;
    }
    nmse = GC_my_mark_stack.top - GC_my_mark_stack.watermark+1;
    if (nmse > 1) {
	GC_my_mark_stack.watermark += nmse/2;
    }
    PSM_mutex_unlock(&GC_my_mark_stack.ml);
    return 0;
}

int GC_get_work_or_terminate(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    int endflag = 0;
    GC_thread envp;
    int nloop = 0;
    int nloop_warning = 100;
    mark_stack_t *from_ms;
    register mse *from_ms_bottom;
    register mse *from_ms_watermark;
    mark_stack_t *to_ms;
    int old_ne;

#if DO_OPT_NSB == 1
    GC_myg_ld.busy_flag = 0;
#else
    PSM_counter_add(&GC_g_ld.busy_count, -1);
    if (GC_g_ld.busy_count == 0) {
	endflag = 1;
	return endflag;
    }
#endif

    GC_LDLOG_PUT(GC_LDC_S, GC_LDK_SEARCH);
    to_ms = &GC_my_mark_stack;
    while (1) {
#ifdef DO_REQ_REMOTE
	if (GC_my_reqbuf.validsize > 0) {
#if DO_OPT_NSB == 1
	    GC_myg_ld.busy_flag = 1;
	    GC_g_ld.flag_modified = 1;
#else
	    PSM_counter_add(&GC_g_ld.busy_count, 1);
#endif
	    endflag = 0;
	    break;
	}
#endif
	/* See other mark stack. */
	for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	    if (GC_l_initialized(envp) == 0) continue;
	    if (GC_l_terminated(envp) != 0) continue;

	    /* Termination Detection */
	    if (nloop > 0 && GC_detect_termination(SGC_MYENV_ARG0)) {
		/* mark phase has finished! */
		endflag = 1;
		break;
	    }

	    if (envp == GC_my_envp) {
		/* Skip self */
		continue;
	    }

	    from_ms = &GC_l_mark_stack(envp);
#if DO_OPT_SEQ == 1
	    /* Skip empty queue(stack). */
	    old_ne = from_ms->watermark - from_ms->bottom;
	    if (old_ne <= 0) {
		continue;
	    }
#endif
#if DO_OPT_SLQ == 1
	    /* This may improve progress. */
	    if (PSM_mutex_trylock(&from_ms->ml) == 0) {
		continue;
	    }
#else
	    MUTEX_LOCK_LOG(&from_ms->ml, GC_LDK_WMQ);
#endif
#if 0 && defined SDSM_TEST
	    {
		char str[256];
		sprintf(str, "[%d:%05x]: get_work_: mutex_locked! ms.ml(%08x) %d loop...\n",
			S_my_pe(), S_my_threadid(),
			&from_ms->ml,
			nloop);
		fprintf(stderr, str);
	    }

#endif
	    from_ms_watermark = from_ms->watermark;
	    from_ms_bottom = from_ms->bottom;
	    /* from_ms->top may be incremented by owner without lock, */
	    /* but that is harmless. */
	    old_ne = from_ms_watermark - from_ms_bottom;

	    if (old_ne <= 0) {
		PSM_mutex_unlock(&from_ms->ml);
		continue;
	    }
	    /* Found work! */
	    PSM_mutex_lock(&to_ms->ml);
	    GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_MOVED, 1);
	    /* Pushing onto to_ms without lock is OK */
#if DO_OPT_SLO == 2
	    {
		int objsz = (from_ms->bottom->mse_descr) & ~DS_TAGS;
		ptr_t p = (ptr_t)from_ms->bottom->mse_start;
		if (objsz >= CUTBYTES_TO_PUSH &&
		    (from_ms->bottom->mse_descr & DS_TAGS) == DS_LENGTH) {
		    /* steal the second half of objects */
		    int remainsz = WORDS_TO_BYTES(BYTES_TO_WORDS(objsz/2));
		    COPY_MSE(to_ms->top+1, from_ms_bottom);
		    (to_ms->top+1)->mse_start = (word *)(p + remainsz);
		    (to_ms->top+1)->mse_descr =
			(((objsz-remainsz+DS_TAGS) & ~DS_TAGS) | DS_LENGTH);
		    from_ms->bottom->mse_start = (word *)p;
		    from_ms->bottom->mse_descr =
			(((remainsz+DS_TAGS) & ~DS_TAGS) | DS_LENGTH);
		    GC_MSLOG_PUT(GC_MSC_STEAL, (to_ms->top+1)->mse_start);
		    to_ms->top++;
		}
		else {
		    COPY_MSE(to_ms->top+1, from_ms_bottom);
		    GC_MSLOG_PUT(GC_MSC_STEAL, (to_ms->top+1)->mse_start);
		    to_ms->top++;
		    from_ms->bottom++;
		}
	    }
#else
	    COPY_MSE(to_ms->top+1, from_ms_bottom);
	    GC_MSLOG_PUT(GC_MSC_STEAL, (to_ms->top+1)->mse_start);
	    to_ms->top++;
	    from_ms->bottom++;
#endif

#if DO_OPT_NSB == 1
	    GC_myg_ld.busy_flag = 1;
	    GC_g_ld.flag_modified = 1;
#else
	    PSM_counter_add(&GC_g_ld.busy_count, 1);
#endif
	    PSM_mutex_unlock(&to_ms->ml);
	    PSM_mutex_unlock(&from_ms->ml);
	    endflag = 0;
	    break;
	}
	if (envp != NULL) break;
	nloop++;
#if 1 && defined SDSM_TEST
	S_usleep((nloop > 10000)? 10000: nloop);
	if (nloop >= nloop_warning) {
	    char str[256];
	    char str2[256];
	    GC_thread envp;
	    sprintf(str, "[%d:%05x]: GC_get_work_or_terminate  %d loop...\n",
		    S_my_pe(), S_my_threadid(),
		    nloop);
	    fprintf(stderr, str);
#if DO_OPT_NSB == 1
	    sprintf(str, "[%d:%05x]: all_finished=%d, dterm_ml=%d, flag_modified=%d\n",
		    S_my_pe(), S_my_threadid(),
		    GC_g_ld.all_finished, GC_g_ld.dterm_ml,
		    GC_g_ld.flag_modified);
	    fprintf(stderr, str);
	    sprintf(str, "[%d:%05x]: busy_flag -> ",
		    S_my_pe(), S_my_threadid());
	    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
		if (GC_l_initialized(envp) == 0) continue;
		if (GC_l_terminated(envp) != 0) continue;
		if (envp == GC_my_envp) {
		    sprintf(str2, "self %d -> ", GC_lg_ld(envp).busy_flag);
		}
		else {
		    sprintf(str2, "%d -> ", GC_lg_ld(envp).busy_flag);
		}
		strcat(str, str2);
	    }
	    strcat(str, "\n");
	    fprintf(stderr, str);
#endif
	    nloop_warning *= 2;
	}
#endif
    }

    GC_LDLOG_PUT(GC_LDC_E, GC_LDK_SEARCH);

    return endflag;
}

#endif /* DO_LOAD_DIST == 5 */

