/*
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */
/* Endo, February 8, 1999 */

#include <stdio.h>
#include "gc_priv.h"

#ifdef ADDRS_SEPARATE
/* need to open, mmap */
#  include <sys/types.h>
#  include <sys/stat.h>
#  include <fcntl.h>
#  include <unistd.h>
#  include <sys/mman.h>
#endif

/* Configuration (valid only in this file) */

#ifndef SILENT
/*#  define SCR_DEBUGPRINT*/
#endif

#ifdef SHM_FIXED
#  define DEFAULT_SHM_SIZE (384*0x100000)
#  define NONTOUCHED_SHM_SIZE(s) (s/8)
#endif

/************************************************ module shm */

#if defined ADDDRS_SEPARATE && defined SDSM_TEST
#  error SDSM_TEST does not support process version!
#endif

#ifdef ADDRS_SEPARATE
/* Multi-Process version (Need mmap to share heap) */

ptr_t GC_sh_shm_alloc(int shm_size, struct shm_handler *hp)
{
    char *fn;
    int fd;
    ptr_t addr;

    fn = "/dev/zero";
#ifdef SCR_DEBUGPRINT
    fprintf(stderr, "Try to open [%s] ..\n", fn);
#endif
    fd = open(fn, O_RDWR | O_CREAT, 0777);
    if (fd < 0) {
	fprintf(stderr, "[GC_sh_shm_alloc] can't open");
	exit(0);
    }
    
    addr = mmap(NULL, shm_size, PROT_READ | PROT_WRITE,
		MAP_SHARED, fd, 0);
#ifdef SCR_DEBUGPRINT
    fprintf(stderr, "[GC_sh_shm_alloc]%08x - %08x\n", addr, addr+shm_size-1);
#endif
    if ((int)addr == -1) {
	fprintf(stderr, "[GC_sh_shm_alloc] can't mmap\n");
	exit(0);
    }

    strcpy(hp->filename, fn);
    hp->filedesc = fd;

    return addr;
}

#else

#ifndef SDSM_TEST
/* Multi-thread version(`malloc'ed region is shared) */
ptr_t GC_sh_shm_alloc(int shm_size, struct shm_handler *hp)
{
    ptr_t addr;
    int align;
    int lsbs;

    align = GETPAGESIZE();
    if (align < HBLKSIZE) {
	align = HBLKSIZE;
    }
    addr = (ptr_t)malloc(shm_size + align);
    if (addr == NULL) {
	fprintf(stderr, "[GC_sh_shm_alloc] can't malloc(%d)\n",
		shm_size);
	return NULL;
    }
    lsbs = (word)addr & (align-1);
    if (lsbs != 0) {
	addr = addr + (align-lsbs);
    }
    return addr;
}

#else /* !SDSM_TEST */

#include <sdsm.h>

ptr_t GC_sh_shm_alloc(int shm_size, struct shm_handler *hp)
{
    ptr_t addr;
    int align;
    int lsbs;

    align = GETPAGESIZE();
    if (align < HBLKSIZE) {
	align = HBLKSIZE;
    }
    addr = (ptr_t)S_malloc(shm_size + align);
    if (addr == NULL) {
	fprintf(stderr, "[GC_sh_shm_alloc] can't S_malloc(%d)\n",
		shm_size);
	return NULL;
    }
    lsbs = (word)addr & (align-1);
    if (lsbs != 0) {
	addr = addr + (align-lsbs);
    }

    return addr;
}

#endif /* !SDSM_TEST */

#endif

/*******************************************************************/
#ifdef SHM_FIXED
/* All of shared region is allocated at once. */
/* GC_sh_shm_alloc() is called only once. */

/* This function must be called only ONCE. NOT MT-safe. */
/* This function must be called before all child(GC/App) processes */
/* are created. */
/* Mmap for shared-heap and allocate GC_arrays. */
int GC_sh_get_mem_init(int shm_size)
{
    char *addr;
    int shm2_size;
    struct shm_handler handler;
    int align;

    if (shm_size <= 0) {
	shm_size = DEFAULT_SHM_SIZE;
    }
    
    if (shm_size <= sizeof(struct _GC_arrays)) {
	fprintf(stderr, "WARNING: Specified size is too small!\n");
	shm_size = sizeof(struct _GC_arrays)+(32*0x100000);
    }

    align = GETPAGESIZE();
    if (align < HBLKSIZE) {
	align = HBLKSIZE;
    }

    shm2_size = NONTOUCHED_SHM_SIZE(shm_size);
    shm2_size = shm2_size / align * align;

    addr = GC_sh_shm_alloc(shm_size, &handler);
    
    /* Allocate GC_arrays first. */
    GC_arraysp = (struct _GC_arrays *)addr;
    BZERO(GC_arraysp, sizeof(struct _GC_arrays));
    
    GC_g_shm.shmaddr = addr;
    GC_g_shm.shmsize = shm_size;
    GC_g_shm.allocptr = addr+sizeof(struct _GC_arrays);
    GC_g_shm.limitptr = addr+shm_size-shm2_size-0x1000;
    GC_g_shm.alloc2ptr = addr+shm_size-shm2_size;
    GC_g_shm.limit2ptr = addr+shm_size-0x1000;

    GC_g_shm.handler = handler;
    PSM_mutex_init(&GC_g_shm.ml);

    GC_g_shm.ospagesize = align;

    return 0;
}

/* MT-safe. */
/* Routines to dynamically allocate data structures that will */
/* never be freed from shared heap. */
/* Return value is page-aligned. */
/* `bytes' should be a multiple of pagesize. */
ptr_t GC_sh_get_mem(bytes)
word bytes;
{
    char *p;
    int lsbs;
    int align = GC_g_shm.ospagesize;

    PSM_mutex_lock(&GC_g_shm.ml);
    lsbs = (word)GC_g_shm.allocptr & (align-1);
    if (lsbs != 0) {
	GC_g_shm.allocptr += align-lsbs;
    }

    if (GC_g_shm.allocptr+bytes >=
	GC_g_shm.limitptr) {
	PSM_mutex_unlock(&GC_g_shm.ml);
	fprintf(stderr, "[GC_sh_get_mem] Currently, cannot expand shmem-heap\n");
	return NULL;
    }
    p = GC_g_shm.allocptr;
    GC_g_shm.allocptr += bytes;
    PSM_mutex_unlock(&GC_g_shm.ml);

    return p;
}

ptr_t GC_sh_get_mem_nontouched(bytes)
word bytes;
{
    char *p;
    int lsbs;
    int align = GC_g_shm.ospagesize;

    PSM_mutex_lock(&GC_g_shm.ml);
    lsbs = (word)GC_g_shm.alloc2ptr & (align-1);
    if (lsbs != 0) {
	GC_g_shm.alloc2ptr += align-lsbs;
    }

    if (GC_g_shm.alloc2ptr+bytes >=
	GC_g_shm.limit2ptr) {
	PSM_mutex_unlock(&GC_g_shm.ml);
	fprintf(stderr, "[GC_sh_get_mem2] Currently, cannot expand shmem-heap\n");
	return NULL;
    }
    p = GC_g_shm.alloc2ptr;
    GC_g_shm.alloc2ptr += bytes;
    PSM_mutex_unlock(&GC_g_shm.ml);

    return p;
}

/* This function must be called only ONCE. NOT MT-safe. */
int GC_sh_get_mem_end()
{
    return 0;
}

#else /* SHM_FIXED */
/* Shared region is allocated on demand. */
/* GC_sh_shm_alloc() is called many times. */

/* This function must be called only ONCE. NOT MT-safe. */
/* Mmap for shared-heap and allocate GC_arrays. */
int GC_sh_get_mem_init(int shm_size)
{
    ptr_t addr;
    int align;
    int lsbs;
    struct shm_handler handler;

    addr = GC_sh_shm_alloc(sizeof(struct _GC_arrays), &handler);
    GC_arraysp = (struct _GC_arrays *)addr;
    
    BZERO(GC_arraysp, sizeof(struct _GC_arrays));
    PSM_mutex_init(&GC_g_shm.ml);
    GC_g_shm.handler = handler;

    return 0;
}

/* MT-safe. */
/* Routines to dynamically allocate data structures that will */
/* never be freed from shared heap. */
/* Return value is page-aligned. */
/* If `bytes' is a multiple of pagesize, */
/* we can avoid wasting memory. */
ptr_t GC_sh_get_mem(bytes)
word bytes;
{
    ptr_t result;
    struct shm_handler handler;
    
    PSM_mutex_lock(&GC_g_shm.ml);
    result = GC_sh_shm_alloc(bytes, &handler);
    PSM_mutex_unlock(&GC_g_shm.ml);
    if (result == NULL) {
	fprintf(stderr, "!!!!GC_sh_get_mem: failed to alloc\n");
    }
#if 0
    {
	char str[256];
	sprintf(str, "GET_MEM(%x) -> %08x\n", bytes, result);
	fprintf(stderr, str);
    }
#endif
    return((ptr_t)result);
}

ptr_t GC_sh_get_mem_untouched(bytes)
word bytes;
{
    return GC_sh_get_mem(bytes);
}

/* This function must be called only ONCE. NOT MT-safe. */
int GC_sh_get_mem_end()
{
    return 0;
}

#endif /* SHM_FIXED */


/************************************************ module scratch_alloc */

/* This function can be called GC/App proc. MT-safe. */
int GC_scratch_alloc_init(scratchp)
struct scratch *scratchp;
{
    PSM_mutex_init(&scratchp->ml);
    scratchp->free_ptr = 0;
    scratchp->end_ptr = 0;
    return 0;
}

/* Routines to dynamically allocate collector data structures that will */
/* never be freed.							 */
/* Both GC process and App process can use this function. */
/* MT-Safe. */
 
ptr_t GC_scratch_generic_alloc(bytes, scratchp)
register word bytes;
struct scratch *scratchp;
{
    register ptr_t result;

    bytes = ((bytes+ALIGNMENT-1)/ALIGNMENT)*ALIGNMENT;

    PSM_mutex_lock(&scratchp->ml);
    if (scratchp->free_ptr + bytes <= scratchp->end_ptr) {
	result = scratchp->free_ptr;
	scratchp->free_ptr += bytes;
	PSM_mutex_unlock(&scratchp->ml);
        return result;
    }
    {
        word bytes_to_get = MINHINCR * HBLKSIZE;
         
        if (bytes_to_get <= bytes) {
	    /* Large region */
	    /* get memory directly */
	    ptr_t p;
	    p = (ptr_t)GET_MEM(bytes);
	    PSM_mutex_unlock(&scratchp->ml);
            return p;
        }
	/* Small region */
        result = (ptr_t)GET_MEM(bytes_to_get);
        if (result == 0) {
	    ptr_t p;
#ifdef PRINTSTATS
	    GC_printf0("Out of memory - trying to allocate less\n");
#endif
	    p = (ptr_t)GET_MEM(bytes);
	    if (p == 0) {
#ifdef PRINTSTATS
                GC_printf0("Out of memory\n");
#endif
	    }
	    PSM_mutex_unlock(&scratchp->ml);
            return p;
        }

	scratchp->free_ptr = result+bytes;
        scratchp->end_ptr = result + bytes_to_get;
	PSM_mutex_unlock(&scratchp->ml);
        return result;
    }
}

/* Allocate from my scratch */
ptr_t GC_scratch_alloc(SGC_MYENV_ARG1 bytes)
SGC_MYENV_ARG_DECL2
register word bytes;
{
    DECLARE_MYENV
    ptr_t result;
    result = GC_scratch_generic_alloc(bytes, &GC_my_scratch);
#if 0
    {
	char str[256];
	sprintf(str, "GC_scratch_alloc(%x) -> %08x\n", bytes, result);
	fprintf(stderr, str);
    }
#endif
    return result;
}

/* Allocate from global scratch */
ptr_t GC_scratch_g_alloc(bytes)
register word bytes;
{
    ptr_t result;
    result = GC_scratch_generic_alloc(bytes, &GC_g_scratch);
#if 0
    {
	char str[256];
	sprintf(str, "GC_scratch_g_alloc(%x) -> %08x\n", bytes, result);
	fprintf(stderr, str);
    }
#endif
    return result;
}
