/*
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */
/* Endo, June 3, 1999 */

/* Parallel reclaim for multi-process/thread */

#include <stdio.h>
#include "gc_priv.h"

static void free_hdr_local GC_PROTO((hdr *hhdr, register SH_HBCHUNK *chunkp)); 

#ifdef DO_PRECLAIM

void GC_hbiter_init_index(index_p)
bottom_index *index_p;
{
    index_p->hbcounter = BOTTOM_SZ/SH_HBITER_NHBLK;
    PSM_mutex_init(&index_p->hbcounter_ml);
    PSM_mutex_init(&index_p->hbiter_ml);
    index_p->hbiter_j = BOTTOM_SZ;
}

int GC_hbiter_start()
{
    bottom_index *index_p;

    PSM_mutex_init(&GC_g_pr.bicounter_ml);
    GC_g_pr.bicounter = 0;

    index_p = GC_all_bottom_indices;
    while (index_p != NULL) {
	GC_hbiter_init_index(index_p);
	GC_g_pr.bicounter ++;
	index_p = index_p->asc_link;
    }

    return 0;
}

SH_HBCHUNK *GC_hbiter_get()
{
    struct hblk *hbp;
    bottom_index *index_p;
    int j;
    SH_HBCHUNK *hbcp;

    while (1) {
	if (GC_g_pr.bicounter == 0) {
	    break;
	}
	
	for (index_p = GC_all_bottom_indices; index_p != NULL;
	     index_p = index_p->asc_link) {
	    register int volatile *pj;
	    pj = &index_p->hbiter_j;
	    if (*pj == 0) continue;
	    if (PSM_mutex_trylock(&index_p->hbiter_ml)) {
		j = *pj;
		if (j == 0) {
		    PSM_mutex_unlock(&index_p->hbiter_ml);
		    continue;
		}
		j -= SH_HBITER_NHBLK;
		hbp = (struct hblk *)(((index_p->key << LOG_BOTTOM_SZ)
				       + (word)j)
				      << LOG_HBLKSIZE);
		hbcp = &index_p->hbchunks[j/SH_HBITER_NHBLK];
		*pj = j;
		PSM_mutex_unlock(&index_p->hbiter_ml);

		if (j == 0) {
		    PSM_mutex_lock(&GC_g_pr.bicounter_ml);
		    GC_g_pr.bicounter --;
		    PSM_mutex_unlock(&GC_g_pr.bicounter_ml);
		}

		hbcp->bottom = hbp;
		hbcp->top = hbp + SH_HBITER_NHBLK;
		hbcp->free_start = NULL;
		hbcp->free_end = NULL;
#ifndef SGC_BALANCED_HDR
		hbcp->hdr_free_start = NULL;
		hbcp->hdr_free_end = NULL;
#endif /* !SGC_BALANCED_HDR */

		return hbcp;
	    }
	}
    }

    return NULL;
}

int GC_hbiter_end(void)
{
    return 0;
}

void GC_dump_chunk(SH_HBCHUNK *chunkp)
{
    struct hblk *p;
    struct hblk *lastp;
    hdr *phdr;
    int n = 0;
    
    lastp = NULL;
    p = chunkp->free_start;
    while (p) {
	n++;
	lastp = p;
	phdr = HDR(p);
	p = phdr->hb_next;
    }
    PSM_mutex_lock(&GC_g_debug.print_ml);
    fprintf(stderr, "This chunk has %d cont-hblks\n", n);
    fprintf(stderr, "Actual last is %016lx, free_end is %016lx\n",
	    (unsigned long)lastp, (unsigned long)chunkp->free_end);
    PSM_mutex_unlock(&GC_g_debug.print_ml);
}

#ifndef SGC_BALANCED_HDR
/* See headers.h */
static void free_hdr_local(hdr *hhdr, SH_HBCHUNK *chunkp)
{
    if (chunkp->hdr_free_start == NULL) {
	chunkp->hdr_free_start = hhdr;
	chunkp->hdr_free_end = hhdr;
	hhdr->hb_next = NULL;
    }
    else {
	hhdr->hb_next = (struct hblk *) chunkp->hdr_free_start;
	chunkp->hdr_free_start = hhdr;
    }
    return;
}
#endif /* !SGC_BALANCED_HDR */

/* Remove the header for block h */
/* The header is linked to chunkp->hdr_free_list (!SGC_BALANCED_HDR). */
/*                         GC_l_hdr_free_lsit (SGC_BALANCED_HDR) */
/* See headers.h */
void GC_remove_header_local(struct hblk *h, SH_HBCHUNK *chunkp)
{
    hdr ** ha;
    
    GET_HDR_ADDR(h, ha);
#ifdef SGC_BALANCED_HDR
    GC_free_balanced_hdr(*ha, h);
#else
    free_hdr_local(*ha, chunkp);
#endif
    *ha = 0;
}

int GC_coalesce_or_link(p1, p2)
struct hblk *p1;
struct hblk *p2;
{
    hdr *phdr1 = HDR(p1);
    if (IS_FORWARDING_ADDR_OR_NIL(phdr1)) {
	fprintf(stderr, "WARNING: no hdr (%ld) for p1=%016lx\n", 
		(long)phdr1, (unsigned long)p1);
	return 0;
    }
    else if ((word)p1+phdr1->hb_sz > (word)p2) {
	fprintf(stderr, "WARNING: wrong p1=%016lx(size %08x), p2=%016lx\n",
		(unsigned long)p1, (unsigned int)phdr1->hb_sz, (unsigned long)p2);
	return 0;
    }
    else if ((word)p1+phdr1->hb_sz == (word)p2) {
	hdr *phdr2 = HDR(p2);
	if (IS_FORWARDING_ADDR_OR_NIL(phdr2)) {
	    fprintf(stderr, "WARNING: no hdr for p2=%016lx\n", (unsigned long)p2);
	}
	phdr1->hb_next = phdr2->hb_next;
	phdr1->hb_sz += phdr2->hb_sz;
	return 1;
    }
    else {
	phdr1->hb_next = p2;
	return 0;
    }
}

void
GC_link_to_hblkfl_local(SGC_MYENV_ARG_DECL1 struct hblk *p)
{
    DECLARE_MYENV
    register hdr *phdr;	/* Header corresponding to p */
    register struct hblk *hbp, *prevhbp;
    register hdr *hhdr, *prevhdr;
    register signed_word size;
    register SH_HBCHUNK *chunkp = GC_myg_pr.chunkp;
    phdr = HDR(p);
    if (IS_FORWARDING_ADDR_OR_NIL(phdr)) {
	fprintf(stderr, "WARNING[link_to]: no hdr for p=%016lx\n", (unsigned long)p);
    }
    size = phdr->hb_sz; /* Must be n*HBLKSIZE */

    prevhbp = 0;
    hbp = chunkp->free_start;
    hhdr = HDR(hbp);

    while( (hbp != 0) && (hbp < p) ) {
	prevhbp = hbp;
	prevhdr = hhdr;
	hbp = hhdr->hb_next;
	hhdr = HDR(hbp);
    }

    if (hbp == 0) {
	chunkp->free_end = p;
	phdr->hb_next = (struct hblk *)NULL;
    }
    
    /* Check for duplicate deallocation in the easy case */
      if ((hbp != 0 && (ptr_t)p + size > (ptr_t)hbp) ||
	  (prevhbp != 0 && (ptr_t)prevhbp + prevhdr->hb_sz > (ptr_t)p)) {
        GC_printf2("[sh..]Duplicate large block deallocation of 0x%lx-0x%lx\n",
        	   (unsigned long) p, (ptr_t)p + size);
        GC_printf2("[sh..]Surrounding free blocks are 0x%lx and 0x%lx\n",
           	   (unsigned long) prevhbp, (unsigned long) hbp);
      }

    /* Coalesce with successor, if possible */
      if( (((word)p)+size) == ((word)hbp) ) {
	phdr->hb_next = hhdr->hb_next;
	phdr->hb_sz += hhdr->hb_sz;
	GC_remove_header_local(hbp, chunkp);
	if (hbp == chunkp->free_end) {
	    chunkp->free_end = p;
	}
      } else {
	phdr->hb_next = hbp;
      }

    if( prevhbp == 0 ) {
	chunkp->free_start = p;
    } else if( (((word)prevhbp) + prevhdr->hb_sz)
      	       == ((word)p) ) {
      /* Coalesce with predecessor */
	prevhdr->hb_next = phdr->hb_next;
	prevhdr->hb_sz += phdr->hb_sz;
	GC_remove_header_local(p, chunkp);
	if (p == chunkp->free_end) {
	    chunkp->free_end = prevhbp;
	}
    } else {
	prevhdr->hb_next = p;
    }

    return;
}

/* Free hblk `p' locally. */
/* `p' is linked to GC_myg_pr.chunkp->free_list. */
void
GC_freehblk_local(SGC_MYENV_ARG1 p)
SGC_MYENV_ARG_DECL2
register struct hblk *p;
{
    register hdr *phdr;	/* Header corresponding to p */
    register signed_word size;

    phdr = HDR(p);
    size = phdr->hb_sz;
    size = HBLKSIZE * OBJ_SZ_TO_BLOCKS(size);
    /* This may change HDRs in other chunks. */
    GC_remove_counts(p, (word)size);
    phdr->hb_sz = size;
    phdr->home_envp = NULL;
    GC_invalidate_map(phdr);

    GC_link_to_hblkfl_local(SGC_MYENV_ARG1 p);
    return;
}


/* Called when a chunk has been swept. */
void
GC_link_my_chunk(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    SH_HBCHUNK *new_chunkp;
    SH_HBCHUNK *res_chunkp;
    bottom_index *index_p;
    int i;

    new_chunkp = GC_myg_pr.chunkp;
    GET_BI(new_chunkp->bottom, index_p);

    PSM_mutex_lock(&index_p->hbcounter_ml);
    index_p->hbcounter--;
    if (index_p->hbcounter > 0) {
	PSM_mutex_unlock(&index_p->hbcounter_ml);
	return;
    }

    PSM_mutex_unlock(&index_p->hbcounter_ml);
    /*** Here, all chunks in this index_p have been swept! */

    res_chunkp = &index_p->hbchunk_res;
    res_chunkp->bottom = 
	(struct hblk *)(((index_p->key << LOG_BOTTOM_SZ) + (word)0)
			<< LOG_HBLKSIZE);
    res_chunkp->top = 
	(struct hblk *)(((index_p->key << LOG_BOTTOM_SZ) + (word)BOTTOM_SZ)
			<< LOG_HBLKSIZE);
    res_chunkp->free_start = NULL;
    res_chunkp->free_end = NULL;
#ifndef SGC_BALANCED_HDR
    res_chunkp->hdr_free_start = NULL;
    res_chunkp->hdr_free_end = NULL;
#endif /* !SGC_BALANCED_HDR */

    for (i = 0; i < BOTTOM_SZ/SH_HBITER_NHBLK; i++) {
	struct hblk *p1, *p2;
	hdr *h1, *h2;

	new_chunkp = &index_p->hbchunks[i];

	/*** Create GC_hblkfreelist (locally) */

	if (res_chunkp->free_start == NULL &&
	    new_chunkp->free_start != NULL) {
	    res_chunkp->free_start = new_chunkp->free_start;
	    res_chunkp->free_end = new_chunkp->free_end;
	}
	else {
	    p1 = res_chunkp->free_end;
	    p2 = new_chunkp->free_start;
	    if (p1 != NULL && p2 != NULL) {
		int coalesched;
		if (p1 >= p2) {
		    fprintf(stderr, "[sh_link_my_] invalid p1=%016lx, p2=%016lx\n",
			    (unsigned long)p1, (unsigned long)p2);
		    fprintf(stderr, "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
		}
		coalesched = GC_coalesce_or_link(p1, p2);
		if (coalesched) {
		    GC_remove_header_local(p2, res_chunkp);
		}
		
		if (!coalesched || p2 != new_chunkp->free_end) {
		    res_chunkp->free_end = new_chunkp->free_end;
		}
	    }
	}

#ifndef SGC_BALANCED_HDR
	/*** Create GC_hdr_free_list (locally) */

	if (res_chunkp->hdr_free_start == NULL &&
	    new_chunkp->hdr_free_start != NULL) {
	    res_chunkp->hdr_free_start = new_chunkp->hdr_free_start;
	    res_chunkp->hdr_free_end = new_chunkp->hdr_free_end;
	}
	else {
	    h1 = res_chunkp->hdr_free_end;
	    h2 = new_chunkp->hdr_free_start;
	    if (h1 != NULL && h2 != NULL) {
		/* link h1 and h2 */
		h1->hb_next = (struct hblk *)h2;
		res_chunkp->hdr_free_end = new_chunkp->hdr_free_end;
	    }
	}
#endif /* !SGC_BALANCED_HDR */
    }

    return;
}

#ifdef FIND_LEAK
void GC_flush_temp_flist(SGC_MYENV_ARG_DECL1 hdr *hhdr)
{
    DECLARE_MYENV
    GC_thread home_envp = hhdr -> home_envp;
    unsigned char kind = hhdr -> hb_obj_kind;
    word sz = hhdr -> hb_sz;
    ptr_t *flist;

    if(GC_myg_pr.freelist_top[kind][sz] == 0){
	return;
    }

    if (home_envp == 0) {
	PSM_mutex_lock(&(hhdr->gfl_ml));
	if(hhdr->gfl == 0){
	    hhdr->gfl = GC_mya_freelist[kind][sz];
	    hhdr->gfl_top = GC_myg_pr.freelist_top[kind][sz];
	    PSM_mutex_lock(&(GC_g_freelist_ml));
	    flist = GC_g_freelist[kind];
	    *((ptr_t *)GC_myg_pr.freelist_top[kind][sz]) = flist[sz];
	    flist[sz] = GC_mya_freelist[kind][sz];
	    PSM_mutex_unlock(&(GC_g_freelist_ml));
	}else{
	    *((ptr_t *)GC_myg_pr.freelist_top[kind][sz]) = 
	      *((ptr_t *)hhdr->gfl_top);
	    *((ptr_t *)hhdr->gfl_top) = GC_mya_freelist[kind][sz];
	}
	GC_mya_freelist[kind][sz] = NULL;
	GC_myg_pr.freelist_top[kind][sz] = NULL;
	PSM_mutex_unlock(&(hhdr->gfl_ml));
	return;
    }

    MUTEX_LOCK_LOG(&(GC_la_pr(home_envp).link_rl_ml),
		   GC_LDK_WRL);
    flist = GC_la_freelist(home_envp)[kind];
    *((ptr_t *)GC_myg_pr.freelist_top[kind][sz]) = flist[sz];
    flist[sz] = GC_mya_freelist[kind][sz];
    GC_mya_freelist[kind][sz] = NULL;
    GC_myg_pr.freelist_top[kind][sz] = NULL;
    PSM_mutex_unlock(&(GC_la_pr(home_envp).link_rl_ml));
    return;
}
#endif /* FIND_LEAK */

void GC_flush_temp_rlist(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    GC_thread home_envp = GC_myg_pr.temp_rlist_home_envp;
    hdr *hhdr;
    struct hblk *hbp;
    struct hblk ***rlists;

    if (GC_myg_pr.temp_rlist_start == NULL) {
	return;
    }

    if (home_envp) {
	/* Link all headers in temp_rlist to home's reclaim_list */
	rlists = GC_la_reclaim_list(home_envp);
	MUTEX_LOCK_LOG(&(GC_la_pr(home_envp).link_rl_ml),
			  GC_LDK_WRL);
    }
    else {
	rlists = GC_g_reclaim_list;
	MUTEX_LOCK_LOG(&(GC_g_pr.link_rl_ml),
			  GC_LDK_WRL);
    }

    hbp = GC_myg_pr.temp_rlist_start;
    while (hbp) {
	int ok;
	int sz;
	struct hblk *nexthbp;

	hhdr = HDR(hbp);
	ok = hhdr->hb_obj_kind;
	sz = hhdr->hb_sz;
	nexthbp = hhdr->hb_next;

	hhdr -> hb_next = rlists[ok][sz];
	rlists[ok][sz] = hbp;

	hbp = nexthbp;
    }

    if (home_envp) {
	/* Link all headers in temp_rlist to home's reclaim_list */
	PSM_mutex_unlock(&(GC_la_pr(home_envp).link_rl_ml));
    }
    else {
	rlists = GC_g_reclaim_list;
	PSM_mutex_unlock(&(GC_g_pr.link_rl_ml));
    }

    GC_myg_pr.temp_rlist_start = NULL;

    return;
}

/* #define DONT_CLEAR_FLOBJ */

void GC_clear_lists_proc(GC_thread envp)
{
    register word size;		/* current object size		*/
    register ptr_t p;	/* pointer to current object	*/
    register struct hblk * h;	/* pointer to block containing *p */
    register hdr * hhdr;
    register int word_no;           /* "index" of *p in *q          */
    ptr_t fl;
    int kind;

#ifndef DONT_CLEAR_FLOBJ
    for (kind = 0; kind < GC_n_kinds; kind++) {
	for (size = 1; size <= MAXOBJSZ; size++) {
	    fl =
#ifdef FIND_LEAK
	      (envp == 0)?
	        GC_g_freelist[kind][size]:
#endif /* FIND_LEAK */
	      GC_la_freelist(envp)[kind][size];
	    for (p= fl; p != 0; p=obj_link(p)){
		h = HBLKPTR(p);
		hhdr = HDR(h);
		if (!hhdr) {
		    GC_err_printf2("SGC error: HDR of %08x is %08x!!!\n",
				   p, hhdr);
		    continue;
		}
		word_no = (((word *)p) - ((word *)h));
#ifdef FIND_LEAK
		set_mark_bit_from_hdr(hhdr, word_no);
#else
/*		GC_finalize_gc(); */
		clear_mark_bit_from_hdr(hhdr, word_no);
#		ifdef GATHERSTATS
		{
		    DECLARE_MYENV
		    DECLARE_MYENV3
		    GC_myg_gc_stat.mem_found -= size;
		}
#		endif
#endif
	    }
	    fl = NULL;
	}
    }
#endif

    if(envp == 0) return; /* global */

    /* Clear reclaim- and free-lists */
    for (kind = 0; kind < GC_n_kinds; kind++) {
	register ptr_t *fop;
	register ptr_t *lim;
	register struct hblk ** rlp;
	register struct hblk ** rlim;
	register struct hblk ** rlist;
        
	rlist = GC_la_reclaim_list(envp)[kind];
	if (rlist == 0) continue;	/* This kind not used.	*/
#ifndef FIND_LEAK
	{
	    lim = &(GC_la_freelist(envp)[kind][MAXOBJSZ+1]);
	    for( fop = GC_la_freelist(envp)[kind];
		 fop < lim; fop++ ) {
		*fop = 0;
	    }
	} /* otherwise free list objects are marked, 	*/
	/* and its safe to leave them			*/
#endif
	rlim = rlist + MAXOBJSZ+1;
	for( rlp = rlist; rlp < rlim; rlp++ ) {
	    *rlp = 0;
	}
    }
    
    return;
}

void GC_clear_lists_local(SGC_MYENV_ARG_DECL0)
{
    DECLARE_MYENV
    if (GC_my_envp == GC_g_gcleader_envp) {
#ifdef FIND_LEAK
	GC_clear_lists_proc(0);
#endif /* FIND_LEAK */
	PSM_mutex_init(&GC_g_pr.link_rl_ml);
    }

    if (GC_np_fixed == 0) {
	/* NP-variable model */
	if (GC_my_envp == GC_g_gcleader_envp) {
	    GC_thread envp;
	    for (envp = GC_g_appproc_envps.head; envp; envp = envp->next) {
		if (GC_l_initialized(envp) == 0) continue;
		if (GC_l_terminated(envp) != 0) continue;
		GC_clear_lists_proc(envp);
		PSM_mutex_init(&GC_la_pr(envp).link_rl_ml);
	    }
	}
    }
    else {
	/* NP-fixed model */
	GC_clear_lists_proc(GC_my_envp);
	PSM_mutex_init(&GC_mya_pr.link_rl_ml);
    }

    return;
}

/* reclaim main rountine. called by each GC process. */
void GC_start_reclaim_local(SGC_MYENV_ARG_DECL0)
{
    DECLARE_MYENV
    int kind;

    /* Reclaim all blocks */
    {
	register SH_HBCHUNK *chunkp;
	register struct hblk *hbp;
	register hdr * hhdr;
	register int j;
	register int n_nohdr;

#ifdef FIND_LEAK
	for (kind = 0; kind < GC_n_kinds; kind++) {
	    for(j = 0; j < MAXOBJSZ + 1; j ++){
		GC_mya_freelist[kind][j] = NULL;
		GC_myg_pr.freelist_top[kind][j] = NULL;
	    }
	}
#endif
	GC_myg_pr.temp_rlist_home_envp = NULL;
	GC_myg_pr.temp_rlist_start = NULL;

	while (1) {
	    chunkp = GC_hbiter_get();
	    n_nohdr = 0;
	    GC_myg_pr.chunkp = chunkp;
	    if (chunkp == NULL) {
		break;
	    }
	    hbp = chunkp->top - 1;
	    j = (chunkp->top - chunkp->bottom) - 1;
	    while (j >= 0) {
		hhdr = HDR(hbp);
		if (!IS_FORWARDING_ADDR_OR_NIL(hhdr)) {
		    if (hhdr->hb_map != GC_invalid_map) {
			GC_reclaim_block(SGC_MYENV_ARG1 hbp,
#ifdef FIND_LEAK
					 TRUE
#else
					 FALSE
#endif
					 );
		    }
		    else {
			/* Already freed, but must be linked to hblkfl again */
			GC_link_to_hblkfl_local(SGC_MYENV_ARG1 hbp);
			GC_MARKLOG_PUT(GC_LOGKIND_N_FREED, 1);
		    }
		    hbp--;
		    j--;
		} else if (hhdr == 0) {
		    hbp--;
		    j--;
		    n_nohdr ++;
		} else {
		    hbp -= (word)(hhdr);
		    j -= (word)(hhdr);
		    n_nohdr ++;
		}
	    }

	    GC_MARKLOG_PUT(GC_LOGKIND_N_NOHDR, n_nohdr);
            GC_link_my_chunk(SGC_MYENV_ARG0);
	    GC_flush_temp_rlist(SGC_MYENV_ARG0);
	}
    }
    
    GC_my_debug_state = 0;

    return;
}

/* Called after all processes finishes local sweeping. */
/* Create GC_hblkfreelist and GC_hdr_free_list. */
int GC_make_hblkfl()
{
    bottom_index *index_p;
    struct hblk *all_free_start;
    struct hblk *all_free_end;
    SH_HBCHUNK *new_chunkp;
    SH_HBCHUNK all_chunk;
    
    all_chunk.free_start = NULL;
    all_chunk.free_end = NULL;
#ifndef SGC_BALANCED_HDR
    all_chunk.hdr_free_start = NULL;
    all_chunk.hdr_free_end = NULL;
#endif /* !SGC_BALANCED_HDR */

    index_p = GC_all_bottom_indices;
    while (index_p != NULL) {
	struct hblk *p1, *p2;
	hdr *h1, *h2;

	new_chunkp = &index_p->hbchunk_res;
	
	/*** Create GC_hblkfreelist */

	if (all_chunk.free_start == NULL &&
	    new_chunkp->free_start != NULL) {
	    all_chunk.free_start = new_chunkp->free_start;
	    all_chunk.free_end = new_chunkp->free_end;
	}
	else {
	    p1 = all_chunk.free_end;
	    p2 = new_chunkp->free_start;
	    if (p1 != NULL && p2 != NULL) {
		int coalesched;
		if (p1 >= p2) {
		    fprintf(stderr, "[sh_make_h_] invalid p1=%016lx, p2=%016lx\n",
			    (unsigned long)p1, (unsigned long)p2);
		}
		coalesched = GC_coalesce_or_link(p1, p2);
		if (coalesched) {
		    GC_remove_header_local(p2, &all_chunk);
		}
		
		if (!coalesched || p2 != new_chunkp->free_end) {
		    all_chunk.free_end = new_chunkp->free_end;
		}
	    }
	}

#ifndef SGC_BALANCED_HDR
	/*** Create GC_hdr_free_list */

	if (all_chunk.hdr_free_start == NULL &&
	    new_chunkp->hdr_free_start != NULL) {
	    all_chunk.hdr_free_start = new_chunkp->hdr_free_start;
	    all_chunk.hdr_free_end = new_chunkp->hdr_free_end;
	}
	else {
	    h1 = all_chunk.hdr_free_end;
	    h2 = new_chunkp->hdr_free_start;
	    if (h1 != NULL && h2 != NULL) {
		/* link h1 and h2 */
		h1->hb_next = (struct hblk *)h2;
		all_chunk.hdr_free_end = new_chunkp->hdr_free_end;
	    }
	}
#endif /* !SGC_BALANCED_HDR */

	/* to next */
	index_p = index_p->asc_link;
    }

    GC_hblkfreelist = all_chunk.free_start;
#ifndef SGC_BALANCED_HDR
    GC_hdr_free_list = all_chunk.hdr_free_start;
#endif

    GC_freehblk_ptr = NULL;
    GC_savhbp = NULL;

    return 0;
}

#endif

